# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import ast
import json
import subprocess
import pkg_resources

from collections.abc import Sequence
from functools import wraps
from timeit import default_timer as timer

from typing import Dict
from configparser import ConfigParser

import click
import requests.exceptions

import dwave.cloud
from dwave.cloud import Client
from dwave.cloud.utils import (
    default_text_input, generate_random_ising_problem,
    datetime_to_timestamp, utcnow, strtrunc, CLIError, set_loglevel,
    get_contrib_packages, user_agent)
from dwave.cloud.coders import bqm_as_file
from dwave.cloud.package_info import __title__, __version__
from dwave.cloud.exceptions import (
    SolverAuthenticationError, InvalidAPIResponseError, UnsupportedSolverError,
    ConfigFileReadError, ConfigFileParseError, SolverNotFoundError, SolverOfflineError,
    RequestTimeout, PollingTimeout)
from dwave.cloud.config import (
    load_profile_from_files, load_config_from_files, get_default_config,
    get_configfile_path, get_default_configfile_path,
    get_configfile_paths)
from dwave.cloud.api.constants import DEFAULT_METADATA_API_ENDPOINT


def enable_logging(ctx, param, value):
    if value and not ctx.resilient_parsing:
        set_loglevel(dwave.cloud.logger, param.name)

def enable_loglevel(ctx, param, value):
    if value and not ctx.resilient_parsing:
        set_loglevel(dwave.cloud.logger, value)

def show_platform(ctx, param, value):
    if value and not ctx.resilient_parsing:
        click.echo(user_agent())
        sys.exit()


def config_file_options(exists=True):
    """Decorate `fn` with `--config-file` and `--profile` options.

    Optionally skip existency check by click with `exists=False`.
    """

    def decorator(fn):
        fn = click.option(
            '--config-file', '-f', default=None, is_eager=True,
            type=click.Path(exists=exists, dir_okay=False),
            help='Configuration file path')(fn)
        fn = click.option(
            '--profile', '-p', default=None,
            help='Connection profile (section) name')(fn)
        return fn

    return decorator


def solver_options(fn):
    """Decorate `fn` with `--client` and `--solver` options."""

    fn = click.option(
        '--client', 'client_type', default=None,
        type=click.Choice(['base', 'qpu', 'sw', 'hybrid'], case_sensitive=False),
        help='Client type used (default: from config)')(fn)
    fn = click.option(
        '--solver', '-s', 'solver_def', default=None,
        help='Feature-based solver filter (default: from config)')(fn)

    return fn


def endpoint_options(fn):
    """Decorate `fn` with `--endpoint` and `--region` options."""

    fn = click.option(
        '--endpoint', default=None, metavar='URL',
        help='Solver API endpoint (default: from config)')(fn)
    # TODO: provide choice for region
    fn = click.option(
        '--region', default=None, metavar='CODE',
        help='Solver API region (default: from config)')(fn)

    return fn


@click.group()
@click.version_option(prog_name=__title__, version=__version__)
@click.option('--debug', is_flag=True, callback=enable_logging,
              help='Enable debug logging.')
@click.option('--trace', is_flag=True, callback=enable_logging,
              help='Enable trace-level debug logging.')
@click.option('--log', 'loglevel', metavar='LEVEL', callback=enable_loglevel,
              help='Set custom numeric or symbolic log level.')
@click.option('--platform', is_flag=True, is_eager=True, callback=show_platform,
              help='Show the platform tags and exit.')
def cli(debug=False, trace=False, loglevel=None, platform=False):
    """D-Wave Cloud Client interactive configuration tool."""


@cli.group()
def config():
    """Create, update or inspect cloud client configuration file(s)."""


@config.command()
@click.option('--system', is_flag=True,
              help='List paths of system-wide configuration files')
@click.option('--user', is_flag=True,
              help='List paths of user-local configuration files')
@click.option('--local', is_flag=True,
              help='List paths of local configuration files')
@click.option('--include-missing', '-m', is_flag=True,
              help='List all examined paths, not only used paths')
def ls(system, user, local, include_missing):
    """List configuration files detected (and/or examined paths)."""

    # default action is to list *all* auto-detected files
    if not (system or user or local):
        system = user = local = True

    for path in get_configfile_paths(system=system, user=user, local=local,
                                     only_existing=not include_missing):
        click.echo(path)


@config.command()
@config_file_options()
def inspect(config_file, profile):
    """Inspect existing configuration/profile."""

    try:
        section = load_profile_from_files(
            [config_file] if config_file else None, profile)

        click.echo("Configuration file: {}".format(config_file if config_file else "auto-detected"))
        click.echo("Profile: {}".format(profile if profile else "auto-detected"))
        click.echo("---")
        for key, val in section.items():
            click.echo("{} = {}".format(key, val))

    except (ValueError, ConfigFileReadError, ConfigFileParseError) as e:
        click.echo(e)


@config.command()
@config_file_options(exists=False)
@click.option('--full', 'ask_full', is_flag=True,
              help='Configure non-essential options (such as endpoint and solver).')
def create(config_file, profile, ask_full):
    """Create or update cloud client configuration file."""

    try:
        _config_create(config_file, profile, ask_full)
    except CLIError as error:
        click.echo(f"Error: {error!s} (code: {error.code})")
        sys.exit(error.code)
    except Exception as error:
        click.echo(f"Unhandled error: {error!s}")
        sys.exit(127)


def _input_config_variables(config: ConfigParser,
                            profile: str,
                            prompts: Dict[str, Dict[str, str]]) -> ConfigParser:
    """Update config variables in place with user-provided values."""

    for var, prompt in prompts.items():
        default_val = config.get(profile, var, fallback=None)
        prompt.setdefault('default', default_val)
        val = default_text_input(**prompt)
        if val:
            val = os.path.expandvars(val)
        if val and val != default_val:
            config.set(profile, var, val)
    return config

def _load_config(config_file: str) -> ConfigParser:
    """Load from config_file, or create new with defaults."""

    try:
        return load_config_from_files([config_file])
    except:
        return get_default_config()

def _write_config(config: ConfigParser, config_file: str):
    """Write config to config_file."""

    config_base = os.path.dirname(config_file)
    if config_base and not os.path.exists(config_base):
        try:
            os.makedirs(config_base)
        except Exception as e:
            raise CLIError(f"Error creating configuration path: {e!r}", code=1)

    try:
        with open(config_file, 'w') as fp:
            config.write(fp)
    except Exception as e:
        raise CLIError(f"Error writing to configuration file: {e!r}", code=2)

def _config_create(config_file, profile, ask_full=False):
    """Full/simplified dwave create flows."""

    if ask_full:
        rs = Client._fetch_available_regions(DEFAULT_METADATA_API_ENDPOINT)
        prompts = dict(
            region=dict(prompt="Solver API region", choices=[r.code for r in rs]),
            endpoint=dict(prompt="Solver API endpoint URL (overwrites 'region')"),
            token=dict(prompt="Authentication token"),
            client=dict(prompt="Client class", choices='base qpu sw hybrid'.split()),
            solver=dict(prompt="Solver"))

    else:
        prompts = dict(
            token=dict(prompt="Authentication token"))

        click.echo("Using the simplified configuration flow.\n"
                   "Try 'dwave config create --full' for more options.\n")

    # resolve config file path
    ask_to_confirm_config_path = not config_file
    if not config_file:
        config_file = get_configfile_path()
        if not config_file:
            config_file = get_default_configfile_path()

    config_file_exists = os.path.exists(config_file)
    verb = "Updating existing" if config_file_exists else "Creating new"
    click.echo(f"{verb} configuration file: {config_file}")

    if ask_full and ask_to_confirm_config_path:
        config_file = default_text_input("Confirm configuration file path", config_file)
        config_file = os.path.expanduser(config_file)

    config = _load_config(config_file)
    default_section = config.default_section

    # resolve profile
    if not profile:
        existing = config.sections()
        if default_section in config:   # not included in sections
            existing.insert(0, default_section)
        if config_file_exists:
            click.echo(f"Available profiles: {', '.join(existing)}")
        default_profile = next(iter(existing))

        _note = " (select existing or create new)" if config_file_exists else ""
        profile = default_text_input(f"Profile{_note}", default_profile)

    verb = "Updating existing" if profile in config else "Creating new"
    click.echo(f"{verb} profile: {profile}")

    if profile != default_section and not config.has_section(profile):
        config.add_section(profile)

    _input_config_variables(config, profile, prompts)

    _write_config(config, config_file)

    click.echo("Configuration saved.")


def _get_client_solver(config, output=None):
    """Helper function to return an instantiated client, and solver, validating
    parameters in the process, while wrapping errors in `CLIError` and using
    `output` writer as a centralized printer.
    """
    if output is None:
        output = click.echo

    # get client
    try:
        client = Client.from_config(**config)
    except Exception as e:
        raise CLIError("Invalid configuration: {}".format(e), code=1)

    config_file = config.get('config_file')
    if config_file:
        output("Using configuration file: {config_file}", config_file=config_file)

    profile = config.get('profile')
    if profile:
        output("Using profile: {profile}", profile=profile)

    output("Using endpoint: {endpoint}", endpoint=client.endpoint)
    output("Using region: {region}", region=client.region)

    # get solver
    try:
        solver = client.get_solver()
    except SolverAuthenticationError:
        raise CLIError("Authentication error. Check credentials in your configuration file.", 2)
    except SolverNotFoundError:
        raise CLIError("Solver not available.", 6)
    except (InvalidAPIResponseError, UnsupportedSolverError):
        raise CLIError("Invalid or unexpected API response.", 3)
    except RequestTimeout:
        raise CLIError("API connection timed out.", 4)
    except requests.exceptions.SSLError as e:
        # we need to handle `ssl.SSLError` wrapped in several exceptions,
        # with differences between py2/3; greping the message is the easiest way
        if 'CERTIFICATE_VERIFY_FAILED' in str(e):
            raise CLIError(
                "Certificate verification failed. Please check that your API endpoint "
                "is correct. If you are connecting to a private or third-party D-Wave "
                "system that uses self-signed certificate(s), please see "
                "https://support.dwavesys.com/hc/en-us/community/posts/360018930954.", 5)
        raise CLIError("Unexpected SSL error while fetching solver: {!r}".format(e), 5)
    except Exception as e:
        raise CLIError("Unexpected error while fetching solver: {!r}".format(e), 5)

    output("Using solver: {solver_id}", solver_id=solver.id)

    return (client, solver)


def _sample(solver, problem, params, output):
    """Blocking sample call with error handling and using custom printer."""

    try:
        response = solver.sample_ising(*problem, **params)
        problem_id = response.wait_id()
        output("Submitted problem ID: {problem_id}", problem_id=problem_id)
        response.result()
    except RequestTimeout:
        raise CLIError("API connection timed out.", 8)
    except PollingTimeout:
        raise CLIError("Polling timeout exceeded.", 9)
    except Exception as e:
        raise CLIError("Sampling error: {!r}".format(e), 10)

    return response


def standardized_output(fn):
    """Decorator that captures `CLIError`s from `fn` and formats output.

    The decorated function (cli command) receives `output()` for info output
    and should raise `CLIError()` (for handled errors) to output error messages.

    The function itself can be invariant to output format and/or error signaling
    mechanism.
    """

    @wraps(fn)
    def wrapped(*args, **kwargs):
        # text/json output taken from callee args
        json_output = kwargs.get('json_output', False)

        now = utcnow()
        info = dict(datetime=now.isoformat(), timestamp=datetime_to_timestamp(now), code=0)

        def output(fmt, maxlen=None, **params):
            info.update(params)
            if not json_output:
                msg = fmt.format(**params)
                if maxlen is not None:
                    msg = strtrunc(msg, maxlen)
                click.echo(msg)

        def flush():
            if json_output:
                click.echo(json.dumps(info))

        try:
            fn(*args, output=output, **kwargs)
        except CLIError as error:
            output("Error: {error} (code: {code})", error=str(error), code=error.code)
            sys.exit(error.code)
        except Exception as error:
            output("Unhandled error: {error}", error=repr(error))
            sys.exit(127)
        finally:
            flush()

    return wrapped


@cli.command()
@config_file_options()
@endpoint_options
@solver_options
@click.option('--sampling-params', '-m', default=None, help='Sampling parameters (JSON encoded)')
@click.option('--request-timeout', default=None, type=float,
              help='Connection and read timeouts (in seconds) for all API requests')
@click.option('--polling-timeout', default=None, type=float,
              help='Problem polling timeout in seconds (time-to-solution timeout)')
@click.option('--label', default='dwave ping', type=str, help='Problem label')
@click.option('--json', 'json_output', default=False, is_flag=True,
              help='JSON output')
@standardized_output
def ping(*, config_file, profile, endpoint, region, client_type, solver_def,
         sampling_params, request_timeout, polling_timeout, label, json_output,
         output):
    """Ping the QPU by submitting a single-qubit problem."""

    # parse params (TODO: move to click validator)
    params = {}
    if sampling_params is not None:
        try:
            params = json.loads(sampling_params)
            assert isinstance(params, dict)
        except:
            raise CLIError("sampling parameters required as JSON-encoded "
                           "map of param names to values", code=99)

    if label:
        params.update(label=label)

    config = dict(
        config_file=config_file, profile=profile,
        endpoint=endpoint, region=region,
        client=client_type, solver=solver_def,
        request_timeout=request_timeout, polling_timeout=polling_timeout)

    t0 = timer()
    client, solver = _get_client_solver(config, output)

    # generate problem
    if hasattr(solver, 'nodes'):
        # structured solver: use the first existing node
        problem = ({min(solver.nodes): 0}, {})
    else:
        # unstructured solver doesn't constrain problem graph
        problem = ({0: 1}, {})

    t1 = timer()
    response = _sample(solver, problem, params, output)

    t2 = timer()
    output("\nWall clock time:")
    output(" * Solver definition fetch: {wallclock_solver_definition:.3f} ms", wallclock_solver_definition=(t1-t0)*1000.0)
    output(" * Problem submit and results fetch: {wallclock_sampling:.3f} ms", wallclock_sampling=(t2-t1)*1000.0)
    output(" * Total: {wallclock_total:.3f} ms", wallclock_total=(t2-t0)*1000.0)
    if response.timing:
        output("\nQPU timing:")
        for component, duration in sorted(response.timing.items()):
            output(" * %(name)s = {%(name)s} us" % {"name": component}, **{component: duration})
    else:
        output("\nQPU timing data not available.")


@cli.command()
@config_file_options()
@endpoint_options
@solver_options
@click.option('--list', '-l', 'list_solvers', default=False, is_flag=True,
              help='Print filtered list of solver names, one per line')
@click.option('--all', '-a', 'list_all', default=False, is_flag=True,
              help='Ignore solver filter (list/print all solvers)')
def solvers(config_file, profile, endpoint, region, client_type, solver_def,
            list_solvers, list_all):
    """Get solver details.

    Solver filter is inherited from environment or the specified configuration
    file and profile.
    """

    if list_all:
        client_type = 'base'
        solver_def = '{}'

    with Client.from_config(
            config_file=config_file, profile=profile,
            endpoint=endpoint, region=region,
            client=client_type, solver=solver_def) as client:

        try:
            solvers = client.get_solvers(**client.default_solver)
        except SolverNotFoundError:
            click.echo("Solver(s) {} not found.".format(solver_def))
            return 1

        if list_solvers:
            for solver in solvers:
                click.echo(solver.id)
            return

        # ~YAML output
        for solver in solvers:
            click.echo("Solver: {}".format(solver.id))
            click.echo("  Parameters:")
            for name, val in sorted(solver.parameters.items()):
                click.echo("    {}: {}".format(name, strtrunc(val) if val else '?'))
            solver.properties.pop('parameters', None)
            click.echo("  Properties:")
            for name, val in sorted(solver.properties.items()):
                click.echo("    {}: {}".format(name, strtrunc(val)))
            click.echo("  Derived properties:")
            for name in sorted(solver.derived_properties):
                click.echo("    {}: {}".format(name, strtrunc(getattr(solver, name))))
            click.echo()


@cli.command()
@config_file_options()
@endpoint_options
@solver_options
@click.option('--biases', '-h', default=None,
              help='List/dict of biases for Ising model problem formulation')
@click.option('--couplings', '-j', default=None,
              help='List/dict of couplings for Ising model problem formulation')
@click.option('--random-problem', '-r', default=False, is_flag=True,
              help='Submit a valid random problem using all qubits')
@click.option('--num-reads', '-n', default=None, type=int,
              help='Number of reads/samples')
@click.option('--label', default='dwave sample', type=str, help='Problem label')
@click.option('--sampling-params', '-m', default=None,
              help='Sampling parameters, JSON encoded')
@click.option('--verbose', '-v', default=False, is_flag=True,
              help='Increase output verbosity')
@click.option('--json', 'json_output', default=False, is_flag=True,
              help='JSON output')
@standardized_output
def sample(*, config_file, profile, endpoint, region, client_type, solver_def,
           biases, couplings, random_problem, num_reads, label, sampling_params,
           verbose, json_output, output):
    """Submit Ising-formulated problem and return samples."""

    # we'll limit max line len in non-verbose mode
    maxlen = None if verbose else 120

    # parse params (TODO: move to click validator)
    params = {}
    if sampling_params is not None:
        try:
            params = json.loads(sampling_params)
            assert isinstance(params, dict)
        except:
            raise CLIError("sampling parameters required as JSON-encoded "
                           "map of param names to values", code=99)

    if num_reads is not None:
        params.update(num_reads=num_reads)

    if label:
        params.update(label=label)

    # TODO: add other params, like timeout?
    config = dict(
        config_file=config_file, profile=profile,
        endpoint=endpoint, region=region,
        client=client_type, solver=solver_def)

    client, solver = _get_client_solver(config, output)

    if random_problem:
        linear, quadratic = generate_random_ising_problem(solver)
    else:
        try:
            linear = ast.literal_eval(biases) if biases else {}
            if isinstance(linear, Sequence):
                linear = dict(enumerate(linear))
        except Exception as e:
            raise CLIError(f"Invalid biases: {e}", code=99)
        try:
            quadratic = ast.literal_eval(couplings) if couplings else {}
        except Exception as e:
            raise CLIError(f"Invalid couplings: {e}", code=99)

    output("Using qubit biases: {linear}", linear=list(linear.items()), maxlen=maxlen)
    output("Using qubit couplings: {quadratic}", quadratic=list(quadratic.items()), maxlen=maxlen)
    output("Sampling parameters: {sampling_params}", sampling_params=params)

    response = _sample(
        solver, problem=(linear, quadratic), params=params, output=output)

    if verbose:
        output("Result: {response!r}", response=response.result())

    output("Samples: {samples!r}", samples=response.samples, maxlen=maxlen)
    output("Occurrences: {num_occurrences!r}", num_occurrences=response.num_occurrences, maxlen=maxlen)
    output("Energies: {energies!r}", energies=response.energies, maxlen=maxlen)


@cli.command()
@config_file_options()
@endpoint_options
@solver_options
@click.option('--problem-id', '-i', default=None,
              help='Problem ID (optional)')
@click.option('--format', '--fmt', default='dimodbqm',
              type=click.Choice(['coo', 'dimodbqm'], case_sensitive=False),
              help='Problem data encoding')
@click.argument('input_file', metavar='FILE', type=click.File('rb'))
def upload(config_file, profile, endpoint, region, client_type, solver_def,
           problem_id, format, input_file):
    """Multipart problem upload with cold restart support."""

    try:
        client = Client.from_config(
            config_file=config_file, profile=profile,
            endpoint=endpoint, region=region,
            client=client_type)
    except Exception as e:
        click.echo("Invalid configuration: {}".format(e))
        return 1
    if config_file:
        click.echo("Using configuration file: {}".format(config_file))
    if profile:
        click.echo("Using profile: {}".format(profile))
    click.echo("Using endpoint: {}".format(client.endpoint))

    click.echo(("Preparing to upload a problem from {!r} "
                "in {!r} format.").format(input_file.name, format))

    if format == 'coo':
        click.echo("Transcoding 'coo' to 'dimodbqm'.")

        try:
            import dimod
        except ImportError: # pragma: no cover
            raise RuntimeError("Can't decode 'coo' format without dimod. "
                               "Re-install the library with 'bqm' support.")

        # note: `BQM.from_coo` doesn't support files opened in binary (yet);
        # fallback to reopen for now
        with open(input_file.name, 'rt') as fp:
            bqm = dimod.BinaryQuadraticModel.from_coo(fp)
            problem_file = bqm_as_file(bqm)

    elif format == 'dimodbqm':
        problem_file = input_file

    click.echo("Uploading...")

    try:
        future = client.upload_problem_encoded(
            problem=problem_file, problem_id=problem_id)
        remote_problem_id = future.result()
    except Exception as e:
        click.echo(e)
        return 2

    click.echo("Upload done. Problem ID: {!r}".format(remote_problem_id))


@cli.command()
@click.option('--list', '-l', 'list_all', default=False, is_flag=True,
              help='List available contrib (non-OSS) packages')
@click.option('--all', '-a', 'install_all', default=False, is_flag=True,
              help='Install all contrib (non-OSS) packages')
@click.option('--update', '-u', 'update_all', default=False, is_flag=True,
              help='Reinstall all installed contrib packages')
@click.option('--accept-license', '--yes', '-y', default=False, is_flag=True,
              help='Accept license(s) without prompting')
@click.option('--verbose', '-v', count=True,
              help='Increase output verbosity (additive, up to 4 times)')
@click.argument('packages', nargs=-1)
def install(list_all, install_all, update_all, accept_license, verbose, packages):
    """Install optional non-open-source Ocean packages."""

    contrib = get_contrib_packages()

    if list_all:
        if verbose:
            # ~YAML output
            for pkg, specs in contrib.items():
                click.echo("Package: {}".format(pkg))
                click.echo("  Title: {}".format(specs['title']))
                click.echo("  Description: {}".format(specs['description']))
                click.echo("  License: {}".format(specs['license']['name']))
                click.echo("  License-URL: {}".format(specs['license']['url']))
                click.echo("  Requires: {}".format(', '.join(specs['requirements'])))
                click.echo()
        else:
            # concise list of available packages
            if contrib:
                click.echo("Available packages: {}.".format(', '.join(contrib.keys())))
            else:
                click.echo("No available packages.")
        return

    if install_all:
        packages = list(contrib)

    if update_all:
        packages = list(filter(_contrib_package_maybe_installed, contrib))

    if not packages:
        click.echo('Nothing to do. Try "dwave install --help".')
        return

    # check all packages requested are registered/available
    for pkg in packages:
        if pkg not in contrib:
            click.echo("Package {!r} not found.".format(pkg))
            return 1

    for pkg in packages:
        _install_contrib_package(pkg, verbose=verbose, prompt=not accept_license)


def _get_dist(dist_spec):
    """Returns `pkg_resources.Distribution` object for matching `dist_spec`,
    which can be given as `pkg_resources.Requirement`, or an unparsed string
    requirement.
    """
    return pkg_resources.get_distribution(dist_spec)


def _contrib_package_maybe_installed(name):
    """Check if contrib package `name` is installed (even partially)."""

    contrib = get_contrib_packages()
    pkg = contrib[name]

    maybe_installed = False
    for req in pkg['requirements']:
        try:
            _get_dist(req)
            maybe_installed = True
        except pkg_resources.VersionConflict:
            # dependency installed, but wrong version
            maybe_installed = True
        except pkg_resources.DistributionNotFound:
            # dependency not installed
            pass

    return maybe_installed


def _install_contrib_package(name, verbose=0, prompt=True):
    """pip install non-oss package `name` from dwave's pypi repo."""

    contrib = get_contrib_packages()
    dwave_contrib_repo = "https://pypi.dwavesys.com/simple"

    assert name in contrib
    pkg = contrib[name]
    title = pkg['title']

    # check if `name` package is already installed
    # right now the only way to check that is to check if all dependencies from
    # requirements are installed
    reinstall = False
    try:
        if all(_get_dist(req) for req in pkg['requirements']):
            click.echo("{} installed and up to date.\n".format(title))
            return
    except pkg_resources.VersionConflict:
        click.echo("{} dependency version mismatch.\n".format(title))
        reinstall = True
    except pkg_resources.DistributionNotFound:
        pass    # dependency not installed, proceed with install

    action = 'Reinstall' if reinstall else 'Install'

    # basic pkg info
    click.echo(title)
    click.echo(pkg['description'])

    # license prompt
    license = pkg['license']
    msgtpl = ("This package is available under the {name} license.\n"
              "The terms of the license are available online: {url}")
    click.echo(msgtpl.format(name=license['name'], url=license['url']))

    if prompt:
        val = default_text_input('{} (y/n)?'.format(action),
                                 default='y', optional=False)
        if val.lower() != 'y':
            click.echo('Skipping: {}.\n'.format(title))
            return

    click.echo('{}ing: {}'.format(action, title))
    for req in pkg['requirements']:

        cmd = [sys.executable, "-m", "pip", "install", req,
               "--extra-index-url", dwave_contrib_repo]
        if verbose > 1:
            cmd.append("-{}".format('v' * (verbose - 1)))

        res = subprocess.run(cmd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)

        if res.returncode or verbose:
            click.echo(res.stdout)

        if res.returncode:
            click.echo('Failed to install {}.\n'.format(title))
            return

    click.echo('Successfully installed {}.\n'.format(title))


@cli.command()
@click.option('--install-all', '--all', '-a', default=False, is_flag=True,
              help='Install all non-open-source packages '\
                   'available and accept licenses without prompting')
@click.option('--full', 'ask_full', default=False, is_flag=True,
              help='Configure non-essential options (such as endpoint and solver).')
@click.option('--verbose', '-v', count=True,
              help='Increase output verbosity (additive, up to 4 times)')
def setup(install_all, ask_full, verbose):
    """Setup optional Ocean packages and configuration file(s).

    Equivalent to running `dwave install [--all]`, followed by
    `dwave config create [--full]`.
    """

    contrib = get_contrib_packages()
    packages = list(contrib)

    if not packages:
        install = False
    elif install_all:
        click.echo("Installing all optional non-open-source packages.\n")
        install = True
    else:
        # The default flow: SDK installed, so some contrib packages registered
        # and `dwave setup` ran without `--all` flag.
        click.echo("Optionally install non-open-source packages and "
                   "configure your environment.\n")
        prompt = "Do you want to select non-open-source packages to install (y/n)?"
        val = default_text_input(prompt, default='y')
        install = val.lower() == 'y'
        click.echo()

    if install:
        for pkg in packages:
            _install_contrib_package(pkg, verbose=verbose, prompt=not install_all)

    click.echo("Creating the D-Wave configuration file.")
    return _config_create(config_file=None, profile=None, ask_full=ask_full)



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import unittest

import requests
import requests_mock

from dwave.cloud.api import exceptions, constants
from dwave.cloud.api.client import DWaveAPIClient, SolverAPIClient, MetadataAPIClient
from dwave.cloud.package_info import __packagename__, __version__


class TestConfig(unittest.TestCase):
    """Session is initiated from config."""

    def test_defaults(self):
        with self.assertRaises(ValueError):
            DWaveAPIClient()

        endpoint = constants.DEFAULT_METADATA_API_ENDPOINT
        client = DWaveAPIClient(endpoint=endpoint)

        defaults = DWaveAPIClient.DEFAULTS.copy()
        defaults.update(endpoint=endpoint)

        self.assertEqual(client.config, defaults)
        self.assertIsInstance(client.session, requests.Session)

        # verify Retry object config
        retry = client.session.get_adapter('https://').max_retries
        conf = DWaveAPIClient.DEFAULTS['retry']
        client.close()
        self.assertEqual(retry.total, conf['total'])

    def test_init(self):
        config = dict(endpoint='https://test.com/path/',
                      token=str(uuid.uuid4()),
                      timeout=1,
                      retry=dict(total=3),
                      headers={'Custom': 'Field 123'},
                      verify=False,
                      proxies={'https': 'http://proxy.com'})

        with DWaveAPIClient(**config) as client:
            session = client.session
            self.assertIsInstance(session, requests.Session)

            self.assertEqual(session.base_url, config['endpoint'])
            self.assertEqual(session.cert, None)
            self.assertEqual(session.headers['X-Auth-Token'], config['token'])
            self.assertEqual(session.headers['Custom'], config['headers']['Custom'])
            self.assertIn(__packagename__, session.headers['User-Agent'])
            self.assertIn(__version__, session.headers['User-Agent'])
            self.assertEqual(session.verify, config['verify'])
            self.assertEqual(session.proxies, config['proxies'])

            # verify Retry object config
            retry = session.get_adapter('https://').max_retries
            self.assertEqual(retry.total, config['retry']['total'])

    def test_sapi_client(self):
        with SolverAPIClient() as client:
            self.assertEqual(client.session.base_url,
                             constants.DEFAULT_SOLVER_API_ENDPOINT)

    def test_metadata_client(self):
        with MetadataAPIClient() as client:
            self.assertEqual(client.session.base_url,
                             constants.DEFAULT_METADATA_API_ENDPOINT)


class TestRequests(unittest.TestCase):

    @requests_mock.Mocker()
    def test_request(self, m):
        """Config options are respected when making requests."""

        config = dict(endpoint='https://test.com/path/',
                      token=str(uuid.uuid4()),
                      headers={'Custom': 'Field 123'})

        auth_headers = {'X-Auth-Token': config['token']}
        data = dict(answer=123)

        m.get(requests_mock.ANY, status_code=401)
        m.get(requests_mock.ANY, status_code=404, request_headers=auth_headers)
        m.get(config['endpoint'], json=data, request_headers=config['headers'])

        with DWaveAPIClient(**config) as client:
            self.assertEqual(client.session.get('').json(), data)

    @requests_mock.Mocker()
    def test_paths(self, m):
        """Path translation works."""

        baseurl = 'https://test.com'
        config = dict(endpoint=baseurl)

        path_a, path_b = 'a', 'b'
        data_a, data_b = dict(answer='a'), dict(answer='b')

        m.get(requests_mock.ANY, status_code=404)
        m.get(f"{baseurl}/{path_a}", json=data_a)
        m.get(f"{baseurl}/{path_b}", json=data_b)

        with DWaveAPIClient(**config) as client:
            self.assertEqual(client.session.get(path_a).json(), data_a)
            self.assertEqual(client.session.get(path_b).json(), data_b)

    @requests_mock.Mocker()
    def test_session_history(self, m):
        """Session history is available."""

        baseurl = 'https://test.com'
        config = dict(endpoint=baseurl, history_size=1)

        m.get(requests_mock.ANY, status_code=404)
        m.get(f"{baseurl}/path", json=dict(data=True))

        with DWaveAPIClient(**config) as client:
            client.session.get('path')
            self.assertEqual(client.session.history[-1].request.path_url, '/path')

            with self.assertRaises(exceptions.ResourceNotFoundError):
                client.session.get('unknown')
                self.assertEqual(client.session.history[-1].exception.error_code, 404)

            client.session.get('/path')
            self.assertEqual(client.session.history[-1].request.path_url, '/path')


class TestResponseParsing(unittest.TestCase):

    @requests_mock.Mocker()
    def test_non_json(self, m):
        """Non-JSON OK response is unexpected."""

        m.get(requests_mock.ANY, text='text', status_code=200)

        with DWaveAPIClient(endpoint='https://mock') as client:

            with self.assertRaises(exceptions.ResourceBadResponseError) as exc:
                client.session.get('test')

    @requests_mock.Mocker()
    def test_structured_error_response(self, m):
        """Error response dict correctly initializes exc."""

        error_msg = "I looked, but couldn't find."
        error_code = 404
        error = dict(error_msg=error_msg, error_code=error_code)

        m.get(requests_mock.ANY, json=error, status_code=error_code)

        with DWaveAPIClient(endpoint='https://mock') as client:

            with self.assertRaisesRegex(exceptions.ResourceNotFoundError, error_msg) as exc:
                client.session.get('test')

                self.assertEqual(exc.error_msg, error_msg)
                self.assertEqual(exc.error_code, error_code)

    @requests_mock.Mocker()
    def test_plain_text_error(self, m):
        """Error messages in plain text/body correctly initialize exc."""

        error_msg = "I looked, but couldn't find."
        error_code = 404

        m.get(requests_mock.ANY, text=error_msg, status_code=error_code)

        with DWaveAPIClient(endpoint='https://mock') as client:

            with self.assertRaisesRegex(exceptions.ResourceNotFoundError, error_msg) as exc:
                client.session.get('test')

                self.assertEqual(exc.error_msg, error_msg)
                self.assertEqual(exc.error_code, error_code)

    @requests_mock.Mocker()
    def test_unknown_errors(self, m):
        """Unknown status code with plain text msg raised as general req exc."""

        error_msg = "I'm a teapot"
        error_code = 418

        m.get(requests_mock.ANY, text=error_msg, status_code=error_code)

        with DWaveAPIClient(endpoint='https://mock') as client:

            with self.assertRaisesRegex(exceptions.RequestError, error_msg) as exc:
                client.session.get('test')

                self.assertEqual(exc.error_msg, error_msg)
                self.assertEqual(exc.error_code, error_code)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings

from dwave.cloud.config import load_config
from dwave.cloud.exceptions import CanceledFutureError, ConfigFileError


# try to load client config needed for live tests on SAPI web service
try:
    # by default, use `test` profile from `tests/dwave.conf`,
    # with secrets (token) read from env
    default_config_path = os.path.join(os.path.dirname(__file__), 'dwave.conf')
    default_config_profile = 'qpu'

    # allow manual override of config file and profile used for tests
    test_config_path = os.getenv('DWAVE_CONFIG_FILE', default_config_path)
    test_config_profile = os.getenv('DWAVE_PROFILE', default_config_profile)

    config = load_config(config_file=test_config_path,
                         profile=test_config_profile)

    # ensure config is complete
    for var in ['token']:
        if not config.get(var):
            raise ValueError("Config incomplete, missing: {!r}".format(var))

except (ConfigFileError, ValueError) as e:
    config = None
    warnings.warn("Skipping live tests due to: {!s}".format(e))



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test problem submission to mock unstructured solvers."""

import io
import json
import unittest
from unittest import mock
from parameterized import parameterized

import numpy

from dwave.cloud.client import Client
from dwave.cloud.solver import (
    BaseUnstructuredSolver, UnstructuredSolver, BQMSolver, CQMSolver, DQMSolver)
from dwave.cloud.concurrency import Present

try:
    import dimod
except ImportError:
    raise unittest.SkipTest("dimod required for unstructured solver tests")


def unstructured_solver_data(problem_type='bqm'):
    return {
        "properties": {
            "supported_problem_types": [problem_type],
            "parameters": {"num_reads": "Number of samples to return."}
        },
        "id": "test-unstructured-solver",
        "description": "A test unstructured solver"
    }

def complete_reply(sampleset, id_="problem-id", type_='bqm', label=None):
    """Reply with the sampleset as a solution."""

    return json.dumps([{
        "status": "COMPLETED",
        "solver": "solver-name",
        "solved_on": "2019-07-31T12:34:56Z",
        "submitted_on": "2019-07-31T12:34:56Z",
        "answer": {
            "format": "bq",
            "data": sampleset.to_serializable()
        },
        "type": type_,
        "id": id_,
        "label": label
    }])

def choose_reply(path, replies):
    """Choose the right response based on the path and make a mock response."""

    if path in replies:
        response = mock.Mock()
        response.status_code = 200
        response.json.side_effect = lambda: json.loads(replies[path])
        return response
    else:
        raise NotImplementedError(path)


class TestUnstructuredSolver(unittest.TestCase):

    def test_sample_bqm_immediate_reply(self):
        """Construction of and sampling from an unstructured BQM solver works."""

        # build a test problem
        bqm = dimod.BQM.from_ising({}, {'ab': 1})

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_problem_id = 'mock-problem-id'
        def mock_upload(self, bqm):
            return Present(result=mock_problem_id)

        # construct a functional solver by mocking client and api response data
        with mock.patch.multiple(Client, create_session=lambda self: session,
                                 upload_problem_encoded=mock_upload):
            with Client(endpoint='endpoint', token='token') as client:
                solver = BQMSolver(client, unstructured_solver_data())

                # make sure this still works
                _ = UnstructuredSolver(client, unstructured_solver_data())

                # direct bqm sampling
                ss = dimod.ExactSolver().sample(bqm)
                ss.info.update(problem_id=mock_problem_id)
                session.post = lambda path, _: choose_reply(
                    path, {'problems/': complete_reply(ss, id_=mock_problem_id)})

                fut = solver.sample_bqm(bqm)
                numpy.testing.assert_array_equal(fut.sampleset, ss)
                numpy.testing.assert_array_equal(fut.samples, ss.record.sample)
                numpy.testing.assert_array_equal(fut.energies, ss.record.energy)
                numpy.testing.assert_array_equal(fut.num_occurrences, ss.record.num_occurrences)

                # submit of pre-uploaded bqm problem
                fut = solver.sample_bqm(mock_problem_id)
                numpy.testing.assert_array_equal(fut.sampleset, ss)
                numpy.testing.assert_array_equal(fut.samples, ss.record.sample)
                numpy.testing.assert_array_equal(fut.energies, ss.record.energy)
                numpy.testing.assert_array_equal(fut.num_occurrences, ss.record.num_occurrences)

                # ising sampling
                lin, quad, _ = bqm.to_ising()
                ss = dimod.ExactSolver().sample_ising(lin, quad)
                ss.info.update(problem_id=mock_problem_id)
                session.post = lambda path, _: choose_reply(
                    path, {'problems/': complete_reply(ss, id_=mock_problem_id)})

                fut = solver.sample_ising(lin, quad)
                numpy.testing.assert_array_equal(fut.sampleset, ss)
                numpy.testing.assert_array_equal(fut.samples, ss.record.sample)
                numpy.testing.assert_array_equal(fut.energies, ss.record.energy)
                numpy.testing.assert_array_equal(fut.num_occurrences, ss.record.num_occurrences)

                # qubo sampling
                qubo, _ = bqm.to_qubo()
                ss = dimod.ExactSolver().sample_qubo(qubo)
                ss.info.update(problem_id=mock_problem_id)
                session.post = lambda path, _: choose_reply(
                    path, {'problems/': complete_reply(ss, id_=mock_problem_id)})

                fut = solver.sample_qubo(qubo)
                numpy.testing.assert_array_equal(fut.sampleset, ss)
                numpy.testing.assert_array_equal(fut.samples, ss.record.sample)
                numpy.testing.assert_array_equal(fut.energies, ss.record.energy)
                numpy.testing.assert_array_equal(fut.num_occurrences, ss.record.num_occurrences)

    def test_sample_cqm_smoke_test(self):
        """Construction of and sampling from an unstructured CQM solver works."""

        # construct a small 3-variable CQM of mixed vartypes
        try:
            import dimod
            mixed = dimod.QM()
            mixed.add_variable('BINARY', 'a')
            mixed.add_variable('SPIN', 'b')
            mixed.add_variable('INTEGER', 'c')
            cqm = dimod.CQM()
            cqm.set_objective(mixed)
            cqm.add_constraint(mixed, rhs=1, sense='==')
        except:
            # dimod or dimod with CQM support not available, so just use a mock
            cqm = mock.Mock()
            cqm.to_file.return_value = io.BytesIO(b'123')

        problem_type = 'cqm'

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_problem_id = 'mock-problem-id'
        def mock_upload(self, bqm):
            return Present(result=mock_problem_id)

        # construct a functional solver by mocking client and api response data
        with mock.patch.multiple(Client, create_session=lambda self: session,
                                 upload_problem_encoded=mock_upload):
            with Client(endpoint='endpoint', token='token') as client:
                solver = CQMSolver(client, unstructured_solver_data(problem_type=problem_type))

                # use bqm for mock response (for now)
                ss = dimod.ExactSolver().sample(dimod.BQM.empty('SPIN'))
                ss.info.update(problem_id=mock_problem_id)
                session.post = lambda path, _: choose_reply(
                    path, {'problems/': complete_reply(ss, id_=mock_problem_id, type_=problem_type)})

                # verify decoding works
                fut = solver.sample_cqm(cqm)
                numpy.testing.assert_array_equal(fut.sampleset, ss)
                numpy.testing.assert_array_equal(fut.problem_type, problem_type)

    def test_sample_dqm_smoke_test(self):
        """Construction of and sampling from an unstructured DQM solver works."""

        try:
            import dimod
            dqm = dimod.DQM()
            dqm.add_variable(5)
            dqm.add_variable(7)
            dqm.set_linear_case(0, 3, 1.5)
            dqm.set_quadratic(0, 1, {(0, 1): 1.5, (3, 4): 1})
        except:
            # dimod or dimod with DQM support not available, so just use a mock
            dqm = mock.Mock()
            dqm.to_file.return_value = io.BytesIO(b'123')

        problem_type = 'dqm'

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_problem_id = 'mock-problem-id'
        def mock_upload(self, bqm):
            return Present(result=mock_problem_id)

        # construct a functional solver by mocking client and api response data
        with mock.patch.multiple(Client, create_session=lambda self: session,
                                 upload_problem_encoded=mock_upload):
            with Client(endpoint='endpoint', token='token') as client:
                solver = DQMSolver(client, unstructured_solver_data(problem_type=problem_type))

                # use bqm for mock response (for now)
                ss = dimod.ExactSolver().sample(dimod.BQM.empty('SPIN'))
                ss.info.update(problem_id=mock_problem_id)
                session.post = lambda path, _: choose_reply(
                    path, {'problems/': complete_reply(ss, id_=mock_problem_id, type_=problem_type)})

                # verify decoding works
                fut = solver.sample_dqm(dqm)
                numpy.testing.assert_array_equal(fut.sampleset, ss)
                numpy.testing.assert_array_equal(fut.problem_type, problem_type)

    def test_upload_failure(self):
        """Submit should gracefully fail if upload as part of submit fails."""

        # build a test problem
        bqm = dimod.BQM.from_ising({}, {'ab': 1})

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_upload_exc = ValueError('error')
        def mock_upload(self, bqm):
            return Present(exception=mock_upload_exc)

        # construct a functional solver by mocking client and api response data
        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:
                with mock.patch.object(BaseUnstructuredSolver, 'upload_problem', mock_upload):
                    solver = UnstructuredSolver(client, unstructured_solver_data())

                    # direct bqm sampling
                    ss = dimod.ExactSolver().sample(bqm)
                    session.post = lambda path, _: choose_reply(
                        path, {'problems/': complete_reply(ss)})

                    fut = solver.sample_bqm(bqm)

                    with self.assertRaises(type(mock_upload_exc)):
                        fut.result()

    def test_many_upload_failures(self):
        """Failure handling in high concurrency mode works correctly."""

        # build a test problem
        bqm = dimod.BQM.from_ising({}, {'ab': 1})

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_upload_exc = ValueError('error')
        def mock_upload(self, bqm):
            return Present(exception=mock_upload_exc)

        # construct a functional solver by mocking client and api response data
        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:
                with mock.patch.object(BaseUnstructuredSolver, 'upload_problem', mock_upload):
                    solver = BQMSolver(client, unstructured_solver_data())

                    futs = [solver.sample_bqm(bqm) for _ in range(100)]

                    for fut in futs:
                        with self.assertRaises(type(mock_upload_exc)):
                            fut.result()


class TestProblemLabel(unittest.TestCase):

    class PrimaryAssertionSatisfied(Exception):
        """Raised by `on_submit_label_verifier` to signal correct label."""

    def on_submit_label_verifier(self, expected_label):
        """Factory for mock Client._submit() that will verify existence, and
        optionally validate label value.
        """

        # replacement for Client._submit()
        def _submit(client, body_data, computation):
            body = json.loads(body_data.result())

            if 'label' not in body:
                if expected_label is None:
                    raise TestProblemLabel.PrimaryAssertionSatisfied
                else:
                    raise AssertionError("label field missing")

            label = body['label']
            if label != expected_label:
                raise AssertionError(
                    "unexpected label value: {!r} != {!r}".format(label, expected_label))

            raise TestProblemLabel.PrimaryAssertionSatisfied

        return _submit

    @parameterized.expand([
        ("undefined", None),
        ("empty", ""),
        ("string", "text label")
    ])
    @mock.patch.object(Client, 'create_session', lambda client: mock.Mock())
    def test_label_is_sent(self, name, label):
        """Problem label is set on problem submit."""

        bqm = dimod.BQM.from_ising({}, {'ab': 1})

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_problem_id = 'mock-problem-id'
        def mock_upload(self, bqm):
            return Present(result=mock_problem_id)

        # construct a functional solver by mocking client and api response data
        with mock.patch.multiple(Client, create_session=lambda self: session,
                                 upload_problem_encoded=mock_upload):
            with Client(endpoint='endpoint', token='token') as client:
                solver = BQMSolver(client, unstructured_solver_data())

                problems = [("sample_ising", (bqm.linear, bqm.quadratic)),
                            ("sample_qubo", (bqm.quadratic,)),
                            ("sample_bqm", (bqm,))]

                for method_name, problem_args in problems:
                    with self.subTest(method_name=method_name):
                        sample = getattr(solver, method_name)

                        with mock.patch.object(Client, '_submit', self.on_submit_label_verifier(label)):

                            with self.assertRaises(self.PrimaryAssertionSatisfied):
                                sample(*problem_args, label=label).result()

    @parameterized.expand([
        ("undefined", None),
        ("empty", ""),
        ("string", "text label")
    ])
    def test_label_is_received(self, name, label):
        """Problem label is set from response in result/sampleset."""

        # build a test problem
        bqm = dimod.BQM.from_ising({}, {'ab': 1})

        # use a global mocked session, so we can modify it on-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_problem_id = 'mock-problem-id'
        def mock_upload(self, bqm):
            return Present(result=mock_problem_id)

        # construct a functional solver by mocking client and api response data
        with mock.patch.multiple(Client, create_session=lambda self: session,
                                 upload_problem_encoded=mock_upload):
            with Client(endpoint='endpoint', token='token') as client:
                solver = BQMSolver(client, unstructured_solver_data())

                # construct mock response
                ss = dimod.ExactSolver().sample(bqm)
                session.post = lambda path, _: choose_reply(
                    path, {'problems/': complete_reply(ss, id_=mock_problem_id, label=label)})

                # sample and verify label
                fut = solver.sample_bqm(bqm, label=label)
                info = fut.sampleset.info
                self.assertIn('problem_id', info)
                if label is not None:
                    self.assertIn('problem_label', info)
                    self.assertEqual(info['problem_label'], label)


class TestSerialization(unittest.TestCase):

    class AssertionSatisfied(Exception):
        """Raised by `on_submit_data_verifier` to signal correct serialization."""

    def on_submit_data_verifier(self, expected_params):
        """Factory for mock Client._submit() that will validate parameter values."""

        # replacement for Client._submit(), called with exact network request data
        def _submit(client, body_data, computation):
            body = json.loads(body_data.result())

            params = body.get('params')
            if params != expected_params:
                raise AssertionError("params don't match")

            raise TestSerialization.AssertionSatisfied

        return _submit

    @parameterized.expand([
        (numpy.bool_(1), True), (numpy.bool8(1), True),
        (numpy.byte(1), 1), (numpy.int8(1), 1),
        (numpy.ubyte(1), 1), (numpy.uint8(1), 1),
        (numpy.short(1), 1), (numpy.int16(1), 1),
        (numpy.ushort(1), 1), (numpy.uint16(1), 1),
        (numpy.intc(1), 1), (numpy.int32(1), 1),
        (numpy.uintc(1), 1), (numpy.uint32(1), 1),
        (numpy.int_(1), 1), (numpy.int32(1), 1),
        (numpy.uint(1), 1), (numpy.uint32(1), 1),
        (numpy.longlong(1), 1), (numpy.int64(1), 1),
        (numpy.ulonglong(1), 1), (numpy.uint64(1), 1),
        (numpy.half(1.0), 1.0), (numpy.float16(1.0), 1.0),
        (numpy.single(1.0), 1.0), (numpy.float32(1.0), 1.0),
        (numpy.double(1.0), 1.0), (numpy.float64(1.0), 1.0),
        (numpy.longdouble(1.0), 1.0)
    ] + ([
        (numpy.float128(1.0), 1.0)      # unavailable on windows
    ] if hasattr(numpy, 'float128') else [
    ]))
    @mock.patch.object(Client, 'create_session', lambda client: mock.Mock())
    def test_params_are_serialized(self, np_val, py_val):
        """Parameters supplied as NumPy types are correctly serialized."""

        user_params = dict(num_reads=np_val)
        expected_params = dict(num_reads=py_val)

        bqm = dimod.BQM.from_ising({}, {'ab': 1})

        # use a global mocked session, so we can modify it on-the-fly
        session = mock.Mock()

        # upload is now part of submit, so we need to mock it
        mock_problem_id = 'mock-problem-id'
        def mock_upload(self, bqm):
            return Present(result=mock_problem_id)

        # construct a functional solver by mocking client and api response data
        with mock.patch.multiple(Client, create_session=lambda self: session,
                                 upload_problem_encoded=mock_upload):
            with Client(endpoint='endpoint', token='token') as client:
                solver = BQMSolver(client, unstructured_solver_data())

                problems = [("sample_ising", (bqm.linear, bqm.quadratic)),
                            ("sample_qubo", (bqm.quadratic,)),
                            ("sample_bqm", (bqm,))]

                for method_name, problem_args in problems:
                    with self.subTest(method_name=method_name, np_val=np_val, py_val=py_val):
                        sample = getattr(solver, method_name)

                        with mock.patch.object(
                                Client, '_submit', self.on_submit_data_verifier(expected_params)):

                            with self.assertRaises(self.AssertionSatisfied):
                                sample(*problem_args, **user_params).result()



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multipart upload helpers."""

import io
import os
import math
import logging
import threading

from abc import abstractmethod
from collections.abc import Sized
from functools import partial

__all__ = ['ChunkedData']

logger = logging.getLogger(__name__)


class Gettable(Sized):
    """Abstract base class for objects that implement standard and efficient
    item getters.

    Concrete subclasses must provide __len__ and __getitem__/getinto.
    """

    __slots__ = ()

    @abstractmethod
    def __getitem__(self, key):     # pragma: no cover
        """Standard item getter, integer and slice indices supported."""
        raise KeyError

    @abstractmethod
    def getinto(self, key, buf):    # pragma: no cover
        """Optimized item getter (without the temporary buffer)."""
        raise NotImplementedError


class GettableBase(Gettable):
    """Base class for sized data containers with a thread-safe efficient item
    getter.

    Subclasses must implement __len__ and getinto.
    """

    def __len__(self):
        raise NotImplementedError

    def _getkey_to_range(self, key):
        """Resolve slice/int key to start-stop range bounds.

        Returns: (start, stop, is_item?)
        """

        if isinstance(key, slice):
            start, stop, stride = key.indices(len(self))
            if stride != 1:
                raise NotImplementedError("stride of 1 required")
            is_item = False
        else:
            try:
                start = int(key)
            except:
                raise TypeError("slice or integral key expected")

            # negative indices wrap around
            if start < 0:
                start %= len(self)

            stop = start + 1
            is_item = True

        return start, stop, is_item

    def getinto(self, key, buf):
        raise NotImplementedError

    def __getitem__(self, key):
        """Fetch a slice of file's content as bytes. For integer index, return
        a single byte value as integer.

        Returns:
            int/:class:`bytes`

        Note:
            Behavior consistent with bytes/bytearray/memoryview.
        """

        start, stop, is_item = self._getkey_to_range(key)
        size = stop - start
        if size <= 0:
            return bytes()

        b = bytearray(size)
        n = self.getinto(key, b)
        del b[n:]

        if is_item:
            # note: for out-of-bounds access this will automatically raise an
            # `IndexError`, as expected, because `n` will be zero
            return b[0]
        else:
            return bytes(b)


class GettableFile(GettableBase):
    """Provide thread-safe memory buffer-like random read access to a file-like
    object via an efficient item getter interface.

    Args:
        fp (:class:`io.BufferedIOBase`/:class:`io.RawIOBase`/binary-file-like):
            A file-like object that supports seek and readinto operations.
            Thread-safety of these operations is not assumed.

        strict (bool, default=True):
            Require file-like object to be a :class:`io.BufferedIOBase` or
            :class:`io.RawIOBase` instance.

    Note:
        :class:`.GettableFile` implements :class:`.Gettable` interface over a
        file-like object that supports buffered binary read access.

    Example:
        Access overlapping segments of a file from multiple threads::

            with open('/path/to/file', 'rb') as fp:   # binary mode, read access
                gf = GettableFile(fp):

                # in thread 1:
                seg = gf[0:10]

                # in thread 2:
                seg = gf[5:15]

    """

    def __init__(self, fp, strict=True):
        if strict:
            valid = lambda f: (
                isinstance(f, (io.BufferedIOBase, io.RawIOBase))
                and f.seekable() and f.readable())
        else:
            valid = lambda f: all([
                hasattr(f, 'readinto'), hasattr(f, 'seek'), hasattr(f, 'tell')])

        if not valid(fp):
            raise TypeError("expected file-like, seekable, readable object")

        # store file size, assuming it won't change
        self._size = fp.seek(0, os.SEEK_END)
        if self._size is None:
            # handle non-python3 and/or non-standard file seek() impl.
            # (like tempfile.SpooledTemporaryFile)
            # note: not thread-safe!
            self._size = fp.tell()

        self._fp = fp

        # multiple threads will be accessing the underlying file 
        self._lock = threading.RLock()

    def __len__(self):
        return self._size

    def getinto(self, key, buf):
        """Copy a slice of file's content into a pre-allocated bytes-like
        object buf. For example, buf might be a `bytearray`.

        Args:
            key (slice/int):
                Source data address coded as a `slice` object or int position.
            buf (bytes-like):
                Target pre-allocated bytes-like object.

        Returns:
            int:
                The number of bytes read (0 for EOF).
        """

        start, stop, _ = self._getkey_to_range(key)

        # empty slice
        if stop <= start:
            return 0

        # copy source[start:stop] => target[0:stop-start]
        size = stop - start
        target = memoryview(buf)[:size]

        # slice is an atomic "seek and read" operation
        with self._lock:
            self._fp.seek(start)
            return self._fp.readinto(target)


class GettableMemory(GettableBase):
    """Provide the :class:`Gettable` interface to a bytes-like object.

    Args:
        buf (bytes/bytearray/memoryview/bytes-like):
            Bytes-like object.

    """

    def __init__(self, buf):
        self._buf = buf

    def __len__(self):
        return len(self._buf)

    def getinto(self, key, buf):
        start, stop, _ = self._getkey_to_range(key)

        # empty slice
        if stop <= start:
            return 0

        # copy source[start:stop] => target[0:stop-start]
        source_view = self._buf[start:stop]
        target_view = memoryview(buf)
        size = min(len(source_view), len(target_view))
        target_view[:size] = source_view[:size]

        return size


class FileView(io.RawIOBase):
    """A raw binary stream subclass with memoryview-like interface to a
    :class:`.GettableBase`-derived object.

    Args:
        raw (:class:`.GettableFile`/:class:`.GettableMemory`/:class:`.GettableBase`-derived):
            A :class:`.GettableBase`-derived object that (essentially) supports
            efficient and thread-safe `getinto` operation.

    Note:
        Although similar to :mod:`mmap` for files, :class:`.FileView` provides:
        - a unified interface to *both* files and memory objects
        - a way to ensure file "seek & read" operation is atomic (thread-safe)
          via the :class:`.Gettable` layer

    Note:
        Use the slice syntax (item getter) to retrieve an isolated file segment
        view (a new instance of :class:`.FileView` that references the same
        underlying data, but supports independent read operations).
    """

    def __init__(self, raw):
        super().__init__()
        self._raw = raw
        self._pos = 0
        self._offset = 0
        self._size = len(raw)

    def seek(self, pos, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            self._pos = pos
        elif whence == os.SEEK_CUR:
            self._pos += pos
        elif whence == os.SEEK_END:
            self._pos = self._size + pos
        else:
            raise ValueError("whence must be one of 'io.SEEK_{SET,CUR,END}'")

        return self._pos

    def tell(self):
        return self._pos

    def readinto(self, b):
        """Read bytes into a pre-allocated bytes-like object b.

        Returns:
            int:
                The number of bytes read.
        """
        start = self._offset + self._pos
        stop = self._offset + self._size
        key = slice(start, stop)
        n = self._raw.getinto(key, b)
        self._pos += n
        return n

    def __len__(self):
        return self._size

    def __getitem__(self, key):
        """Return memoryview-like slice: byte value for an integer index,
        sub-FileView for a slice index.
        """

        # slice key
        if isinstance(key, slice):
            start, stop, stride = key.indices(len(self))
            if stride != 1:
                raise NotImplementedError("stride of 1 required")

            view = FileView(self._raw)
            view._offset += start
            view._size = stop - start
            return view

        # integer key
        try:
            start = int(key)
        except:
            raise TypeError("slice or integer key expected")

        # negative indices wrap around
        if start < 0:
            start %= len(self)

        return self._raw[self._offset + start]


class ChunkedData(object):
    """Unifying and performant streaming file-like interface to (large problem)
    data chunks.

    Handles in-file and in-memory data. Non-seekable streams are not yet
    supported. Provides access to chunk data via file-like interface.

    Args:
        data (bytes-like/file-like):
            Encoded problem data, in-memory or in-file (opened for binary read).

        chunk_size (int):
            Chunk size in bytes.

    """

    def _thread_safe_data_view(self, data):
        # convenience string handler
        if isinstance(data, str):
            data = data.encode('ascii')

        if isinstance(data, (bytes, bytearray)):
            # note: `BytesIO` (in py3.5+) will use buffer protocol (reuse data
            # buffer) only for `bytes` objects, and make a copy otherwise!

            # use non-locking memory view over data buffer
            return FileView(GettableMemory(data))

        # use locking file view if possible
        try:
            return FileView(GettableFile(data, strict=True))
        except TypeError:
            logger.debug("data does not conform to strict file-like requirements")

        # TODO: use stream view if possible

        # fallback to a less strict check of file-like's capabilities,
        # accepting we might fail later
        try:
            return FileView(GettableFile(data, strict=False))
        except TypeError:
            logger.debug("data does not conform to loose file-like requirements")

        raise TypeError("bytes/str/file-like data required")

    def __init__(self, data, chunk_size):
        self.data = data
        self.view = self._thread_safe_data_view(data)

        if chunk_size <= 0:
            raise ValueError("positive integer required for chunk size")
        self.chunk_size = int(chunk_size)

    @property
    def total_size(self):
        """Total data size, in bytes."""
        return len(self.view)

    @property
    def num_chunks(self):
        """Total number of chunks."""
        return math.ceil(self.total_size / self.chunk_size)

    def __len__(self):
        return self.num_chunks

    def chunk(self, idx):
        """Return binary file-like object for a specified data chunk.

        Args:
            idx (int):
                Zero-based chunk index.

        Returns:
            :class:`.FileView`

        """

        start = idx * self.chunk_size
        stop = start + self.chunk_size
        return self.view[start:stop]

    def __iter__(self):
        for idx in range(len(self)):
            yield self.chunk(idx)

    def generators(self):
        """Iterator of (immutable) chunk generators."""

        for idx in range(len(self)):
            yield partial(self.chunk, idx=idx)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Try to load solver data from mock servers."""

import json
import unittest

import requests_mock

from dwave.cloud.client import Client
from dwave.cloud.client.qpu import Client as QPUClient
from dwave.cloud.client.sw import Client as SoftwareClient
from dwave.cloud.client.hybrid import Client as HybridClient
from dwave.cloud.solver import Solver
from dwave.cloud.exceptions import *


url = 'https://dwavesys.com'
token = 'abc123abc123abc123abc123abc123abc123'
solver1_name = 'first_solver'
solver2_name = 'second_solver'

bad_url = 'https://not-a-subdomain.dwavesys.com'
bad_token = '------------------------'


def structured_solver_data(id_, cat='qpu', incomplete=False):
    """Return data dict describing a single solver."""
    obj = {
        "properties": {
            "supported_problem_types": ["qubo", "ising"],
            "qubits": [1, 2, 3],
            "couplers": [[1, 2], [1, 3], [2, 3]],
            "num_qubits": 3,
            "category": cat,
            "parameters": {"num_reads": "Number of samples to return."}
        },
        "id": id_,
        "description": "A test solver",
        "status": "ONLINE"
    }

    if incomplete:
        del obj['properties']['supported_problem_types']

    return obj

def solver_object(id_, cat='qpu', incomplete=False):
    return Solver(client=None, data=structured_solver_data(id_, cat, incomplete))


# Define the endpoints
solver1_url = '{}/solvers/remote/{}/'.format(url, solver1_name)
solver2_url = '{}/solvers/remote/{}/'.format(url, solver2_name)
all_solver_url = '{}/solvers/remote/'.format(url)


def setup_server(m):
    """Add endpoints to the server."""

    solver1_data = structured_solver_data(solver1_name)
    solver2_data = structured_solver_data(solver2_name)
    all_solver_data = [solver1_data, solver2_data]

    # Setup the server
    valid_token_headers = {'X-Auth-Token': token}
    invalid_token_headers = {'X-Auth-Token': bad_token}

    m.get(requests_mock.ANY, status_code=404)
    m.get(requests_mock.ANY, status_code=401, request_headers=invalid_token_headers)

    m.get(solver1_url, json=solver1_data, request_headers=valid_token_headers)
    m.get(solver2_url, json=solver2_data, request_headers=valid_token_headers)
    m.get(all_solver_url, json=all_solver_data, request_headers=valid_token_headers)


class MockConnectivityTests(unittest.TestCase):
    """Test connecting some related failure modes."""

    def test_bad_url(self):
        """Connect with a bad URL."""
        with requests_mock.Mocker() as m:
            setup_server(m)
            with self.assertRaises(SAPIError) as err:
                with Client(endpoint=bad_url, token=token) as client:
                    client.get_solvers()
            # TODO: fix when exceptions/sapi call generalized
            self.assertEqual(err.exception.error_code, 404)

    def test_bad_token(self):
        """Connect with a bad token."""
        with requests_mock.Mocker() as m:
            setup_server(m)
            with self.assertRaises(SolverAuthenticationError) as err:
                with Client(endpoint=url, token=bad_token) as client:
                    client.get_solvers()

    def test_good_connection(self):
        """Connect with a valid URL and token."""
        with requests_mock.mock() as m:
            setup_server(m)
            with Client(endpoint=url, token=token) as client:
                self.assertTrue(len(client.get_solvers()) > 0)


class MockSolverLoading(unittest.TestCase):
    """Test loading solvers in a few different configurations.

    Note:
        A mock server does not test authentication.

    Expect three responses from the server for /solvers/*:
        - A single solver when a single solver is requested
        - A list of single solver objects when all solvers are requested
        - A 404 error when the requested solver does not exist
    An additional condition to test for is that the server may have configured
    a given solver such that it does not provide all the required information
    about it.
    """

    def test_load_solver(self):
        """Load a single solver."""
        with requests_mock.mock() as m:
            setup_server(m)

            # test default, cached solver get
            with Client(endpoint=url, token=token) as client:
                # fetch solver not present in cache
                solver = client.get_solver(solver1_name)
                self.assertEqual(solver.id, solver1_name)

                # modify cached solver and re-fetch it
                solver.id = 'different-solver'
                # cached solver name doesn't match, so it won't be returned
                with self.assertRaises(SolverError):
                    client.get_solver(solver1_name, refresh=False)
                # cache is refreshed?
                self.assertEqual(client.get_solver(solver1_name, refresh=True).id, solver1_name)

    def test_load_all_solvers(self):
        """Load the list of solver names."""

        def spoof_cache(client, clear_val=False, clear_expires=False):
            cache = client._fetch_solvers._cached.cache
            for args in cache:
                if clear_val:
                    cache[args]['val'] = []
                if clear_expires:
                    cache[args]['expires'] = 0

        with requests_mock.mock() as m:
            setup_server(m)

            # test default case, fetch all solvers for the first time
            with Client(endpoint=url, token=token) as client:
                solvers = client.get_solvers()

                self.assertEqual(len(solvers), 2)

                # test default refresh
                spoof_cache(client, clear_expires=True)
                self.assertEqual(len(client.get_solvers()), 2)      # should refresh

                # test no refresh
                spoof_cache(client, clear_val=True)
                self.assertEqual(len(client.get_solvers(refresh=False)), 0)     # should not refresh

                # test refresh
                self.assertEqual(len(client.get_solvers(refresh=True)), 2)      # should refresh

    def test_load_missing_solver(self):
        """Try to load a solver that does not exist."""
        with requests_mock.mock() as m:
            m.get(requests_mock.ANY, status_code=404)
            with Client(endpoint=url, token=token) as client:
                with self.assertRaises(SolverNotFoundError):
                    client.get_solver(solver1_name)

    def test_load_solver_missing_data(self):
        """Try to load a solver that has incomplete data."""
        with requests_mock.mock() as m:
            m.get(solver1_url, json=structured_solver_data(solver1_name, incomplete=True))
            with Client(endpoint=url, token=token) as client:
                with self.assertRaises(SolverNotFoundError):
                    client.get_solver(solver1_name)

    def test_load_solver_broken_response(self):
        """Try to load a solver for which the server has returned a truncated response."""
        with requests_mock.mock() as m:
            body = json.dumps(structured_solver_data(solver1_name))
            m.get(solver1_url, text=body[0:len(body)//2])
            with Client(endpoint=url, token=token) as client:
                with self.assertRaises(InvalidAPIResponseError):
                    client.get_solver(solver1_name)

    def test_get_solver_reproducible(self):
        """get_solver should return same solver (assuming cache hasn't changed)"""

        with requests_mock.mock() as m:
            setup_server(m)

            # prefer solvers with longer name: that's our second solver
            defaults = dict(solver=dict(order_by=lambda s: -len(s.id)))

            with Client(endpoint=url, token=token, defaults=defaults) as client:
                solver = client.get_solver()
                self.assertEqual(solver.id, solver2_name)

                solver = client.get_solver()
                self.assertEqual(solver.id, solver2_name)

    def test_solver_filtering_in_client(self):
        # base client
        self.assertTrue(Client.is_solver_handled(solver_object('test', 'qpu')))
        self.assertTrue(Client.is_solver_handled(solver_object('test', 'software')))
        self.assertTrue(Client.is_solver_handled(solver_object('test', 'hybrid')))
        self.assertTrue(Client.is_solver_handled(solver_object('test', 'whatever')))
        self.assertTrue(Client.is_solver_handled(None))
        # qpu client
        self.assertTrue(QPUClient.is_solver_handled(solver_object('test', 'qpu')))
        self.assertFalse(QPUClient.is_solver_handled(solver_object('test', 'software')))
        self.assertFalse(QPUClient.is_solver_handled(solver_object('test', 'hybrid')))
        self.assertFalse(QPUClient.is_solver_handled(solver_object('test', 'whatever')))
        self.assertFalse(QPUClient.is_solver_handled(None))
        # sw client
        self.assertFalse(SoftwareClient.is_solver_handled(solver_object('test', 'qpu')))
        self.assertTrue(SoftwareClient.is_solver_handled(solver_object('test', 'software')))
        self.assertFalse(SoftwareClient.is_solver_handled(solver_object('test', 'hybrid')))
        self.assertFalse(SoftwareClient.is_solver_handled(solver_object('test', 'whatever')))
        self.assertFalse(SoftwareClient.is_solver_handled(None))
        # hybrid client
        self.assertFalse(HybridClient.is_solver_handled(solver_object('test', 'qpu')))
        self.assertFalse(HybridClient.is_solver_handled(solver_object('test', 'software')))
        self.assertTrue(HybridClient.is_solver_handled(solver_object('test', 'hybrid')))
        self.assertFalse(HybridClient.is_solver_handled(solver_object('test', 'whatever')))
        self.assertFalse(HybridClient.is_solver_handled(None))

    def test_solver_feature_properties(self):
        self.assertTrue(solver_object('solver', 'qpu').qpu)
        self.assertTrue(solver_object('solver', 'QPU').qpu)
        self.assertFalse(solver_object('solver', 'qpu').software)
        self.assertFalse(solver_object('solver', 'qpu').hybrid)
        self.assertFalse(solver_object('solver', 'software').qpu)
        self.assertTrue(solver_object('solver', 'software').software)
        self.assertFalse(solver_object('solver', 'software').hybrid)
        self.assertTrue(solver_object('solver', 'hybrid').hybrid)
        self.assertFalse(solver_object('solver', 'hybrid').qpu)
        self.assertFalse(solver_object('solver', 'hybrid').software)

        self.assertFalse(solver_object('solver').is_vfyc)
        self.assertEqual(solver_object('solver').num_qubits, 3)
        self.assertFalse(solver_object('solver').has_flux_biases)

        # test .num_qubits vs .num_actual_qubits
        data = structured_solver_data('test')
        data['properties']['num_qubits'] = 7
        solver = Solver(None, data)
        self.assertEqual(solver.num_qubits, 7)
        self.assertEqual(solver.num_active_qubits, 3)

        # test .is_vfyc
        data = structured_solver_data('test')
        data['properties']['vfyc'] = 'error'
        self.assertFalse(Solver(None, data).is_vfyc)
        data['properties']['vfyc'] = True
        self.assertTrue(Solver(None, data).is_vfyc)

        # test .has_flux_biases
        self.assertFalse(Solver(None, data).has_flux_biases)
        data['properties']['parameters']['flux_biases'] = '...'
        self.assertTrue(Solver(None, data).has_flux_biases)

        # test .has_anneal_schedule
        self.assertFalse(Solver(None, data).has_anneal_schedule)
        data['properties']['parameters']['anneal_schedule'] = '...'
        self.assertTrue(Solver(None, data).has_anneal_schedule)

        # test `.online` property
        self.assertTrue(solver_object('solver').online)
        data = structured_solver_data('test')
        data['status'] = 'offline'
        self.assertFalse(Solver(None, data).online)
        del data['status']
        self.assertTrue(Solver(None, data).online)

    # Test fallback for legacy solvers without the `category` property
    # TODO: remove when all production solvers are updated
    def test_solver_with_category_missing(self):

        # client type filtering support
        self.assertTrue(QPUClient.is_solver_handled(solver_object('solver', cat='')))
        self.assertTrue(SoftwareClient.is_solver_handled(solver_object('c4-sw_x', cat='')))
        self.assertTrue(HybridClient.is_solver_handled(solver_object('hybrid_x', cat='')))

        # derived properties are correct
        self.assertTrue(solver_object('solver', cat='').qpu)
        self.assertFalse(solver_object('solver', cat='').software)
        self.assertFalse(solver_object('solver', cat='').hybrid)
        self.assertFalse(solver_object('c4-sw_x', cat='').qpu)
        self.assertTrue(solver_object('c4-sw_x', cat='').software)
        self.assertFalse(solver_object('c4-sw_x', cat='').hybrid)
        self.assertFalse(solver_object('hybrid_x', cat='').qpu)
        self.assertFalse(solver_object('hybrid_x', cat='').software)
        self.assertTrue(solver_object('hybrid_x', cat='').hybrid)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import logging
import importlib

from dwave.cloud.client import Client
from dwave.cloud.solver import Solver
from dwave.cloud.computation import Future
from dwave.cloud.utils import set_loglevel

__all__ = ['Client', 'Solver', 'Future']


class FilteredSecretsFormatter(logging.Formatter):
    """Logging formatter that filters out secrets (like Solver API tokens).

    Note: we assume, for easier disambiguation, a secret/token is prefixed with
    a short alphanumeric string, and comprises 40 or more hex digits.
    """

    _SECRETS_PATTERN = re.compile(
        r'\b([0-9A-Za-z]{2,4})-([0-9A-Fa-f]{3})([0-9A-Fa-f]{34,})([0-9A-Fa-f]{3})\b')

    def format(self, record):
        output = super().format(record)
        filtered = re.sub(self._SECRETS_PATTERN, r'\1-\2...\4', output)
        return filtered

# configure logger `dwave.cloud` root logger, inherited in submodules
# (write level warning+ to stderr, include timestamp/module/level)
_formatter = FilteredSecretsFormatter('%(asctime)s %(name)s %(levelname)s %(threadName)s %(message)s')
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)

# expose the root logger to simplify access; for example:
# `dwave.cloud.logger.setLevel(logging.DEBUG)`
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
logger.addHandler(_handler)


# add TRACE log level and Logger.trace() method
logging.TRACE = 5
logging.addLevelName(logging.TRACE, "TRACE")

def _trace(logger, message, *args, **kwargs):
    if logger.isEnabledFor(logging.TRACE):
        logger._log(logging.TRACE, message, args, **kwargs)

logging.Logger.trace = _trace


# apply DWAVE_LOG_LEVEL
def _apply_loglevel_from_env(logger):
    set_loglevel(logger, os.getenv('DWAVE_LOG_LEVEL'))

_apply_loglevel_from_env(logger)


# alias dwave.cloud.client.{qpu,sw,hybrid} as dwave.cloud.*
def _alias_old_client_submodules():
    for name in ('qpu', 'sw', 'hybrid'):
        # note: create both module and local attribute
        globals()[name] = sys.modules['dwave.cloud.{}'.format(name)] = \
            importlib.import_module('dwave.cloud.client.{}'.format(name))

_alias_old_client_submodules()



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import base64

from dwave.cloud.utils import uniform_get, active_qubits

__all__ = [
    'encode_problem_as_qp', 'decode_qp', 'decode_qp_numpy',
    'encode_problem_as_bq', 'decode_bq',
    'encode_problem_as_ref',
    'bqm_as_file',
]


def encode_problem_as_qp(solver, linear, quadratic, offset=0,
                         undirected_biases=False):
    """Encode the binary quadratic problem for submission to a given solver,
    using the `qp` format for data.

    Args:
        solver (:class:`dwave.cloud.solver.Solver`):
            The solver used.

        linear (dict[variable, bias]/list[variable, bias]):
            Linear terms of the model.

        quadratic (dict[(variable, variable), bias]):
            Quadratic terms of the model.

        offset (number, default=0):
            Constant offset applied to the model.

        undirected_biases (boolean, default=False):
            Are (quadratic) biases specified on undirected edges?

    Returns:
        Encoded submission dictionary.
    """
    active = active_qubits(linear, quadratic)

    # Encode linear terms. The coefficients of the linear terms of the objective
    # are encoded as an array of little endian 64 bit doubles.
    # This array is then base64 encoded into a string safe for json.
    # The order of the terms is determined by the _encoding_qubits property
    # specified by the server.
    # Note: only active qubits are coded with double, inactive with NaN
    nan = float('nan')
    lin = [uniform_get(linear, qubit, 0 if qubit in active else nan)
           for qubit in solver._encoding_qubits]

    lin = base64.b64encode(struct.pack('<' + ('d' * len(lin)), *lin))

    # Encode the coefficients of the quadratic terms of the objective
    # in the same manner as the linear terms, in the order given by the
    # _encoding_couplers property, discarding tailing zero couplings
    if undirected_biases:
        # quadratic biases are given in a triangular or symmetric matrix
        quad = [quadratic.get((q1,q2), quadratic.get((q2,q1), 0))
                for (q1,q2) in solver._encoding_couplers
                if q1 in active and q2 in active]
    else:
        # quadratic biases are defined on directed edges, conflate with sum
        quad = [quadratic.get((q1,q2), 0) + quadratic.get((q2,q1), 0)
                for (q1,q2) in solver._encoding_couplers
                if q1 in active and q2 in active]

    quad = base64.b64encode(struct.pack('<' + ('d' * len(quad)), *quad))

    # The name for this encoding is 'qp' and is explicitly included in the
    # message for easier extension in the future.
    return {
        'format': 'qp',
        'lin': lin.decode('utf-8'),
        'quad': quad.decode('utf-8'),
        'offset': offset
    }


def decode_qp(msg):
    """Decode SAPI response that uses `qp` format, without numpy.

    The 'qp' format is the current encoding used for problems and samples.
    In this encoding the reply is generally json, but the samples, energy,
    and histogram data (the occurrence count of each solution), are all
    base64 encoded arrays.
    """
    # Decode the simple buffers
    result = msg['answer']
    result['active_variables'] = _decode_ints(result['active_variables'])
    active_variables = result['active_variables']
    if 'num_occurrences' in result:
        result['num_occurrences'] = _decode_ints(result['num_occurrences'])
    result['energies'] = _decode_doubles(result['energies'])

    # adjust energies by offset (in future this might be handled by SAPI)
    offset = result.setdefault('offset', 0)
    if offset:
        result['energies'] = [en + offset for en in result['energies']]

    # Measure out the size of the binary solution data
    num_solutions = len(result['energies'])
    num_variables = len(result['active_variables'])
    solution_bytes = -(-num_variables // 8)  # equivalent to int(math.ceil(num_variables / 8.))
    total_variables = result['num_variables']

    # Figure out the null value for output
    default = 3 if msg['type'] == 'qubo' else 0

    # Decode the solutions, which will be byte aligned in binary format
    binary = base64.b64decode(result['solutions'])
    solutions = []
    for solution_index in range(num_solutions):
        # Grab the section of the buffer related to the current
        buffer_index = solution_index * solution_bytes
        solution_buffer = binary[buffer_index:buffer_index + solution_bytes]
        bytes = struct.unpack('B' * solution_bytes, solution_buffer)

        # Assume None values
        solution = [default] * total_variables
        index = 0
        for byte in bytes:
            # Parse each byte and read how ever many bits can be
            values = _decode_byte(byte)
            for _ in range(min(8, len(active_variables) - index)):
                i = active_variables[index]
                index += 1
                solution[i] = values.pop()

        # Switch to the right variable space
        if msg['type'] == 'ising':
            values = {0: -1, 1: 1}
            solution = [values.get(v, default) for v in solution]
        solutions.append(solution)

    result['solutions'] = solutions

    # include problem type
    if 'type' in msg:
        result['problem_type'] = msg['type']

    return result


def _decode_byte(byte):
    """Helper for decode_qp, turns a single byte into a list of bits.

    Args:
        byte (int):
            Byte to be decoded.

    Returns:
        List of bits corresponding to byte.
    """
    bits = []
    for _ in range(8):
        bits.append(byte & 1)
        byte >>= 1
    return bits


def _decode_ints(message):
    """Helper for decode_qp, decodes an int array.

    The int array is stored as little endian 32 bit integers.
    The array has then been base64 encoded. Since we are decoding we do these
    steps in reverse.

    Args:
        message (str):
            The int array, base-64 encoded.

    Returns:
        Decoded double array.
    """
    binary = base64.b64decode(message)
    return struct.unpack('<' + ('i' * (len(binary) // 4)), binary)


def _decode_doubles(message):
    """Helper for decode_qp, decodes a double array.

    The double array is stored as little endian 64 bit doubles.
    The array has then been base64 encoded. Since we are decoding we do these
    steps in reverse.

    Args:
        message (str):
            The double array, base-64 encoded.

    Returns:
        Decoded double array.
    """
    binary = base64.b64decode(message)
    return struct.unpack('<' + ('d' * (len(binary) // 8)), binary)


def decode_qp_numpy(msg, return_matrix=True):
    """Decode SAPI response, results in a `qp` format, explicitly using numpy.
    If numpy is not installed, the method will fail.

    To use numpy for decoding, but return the results as lists (instead of
    numpy matrices), set `return_matrix=False`.
    """
    import numpy as np

    result = msg['answer']

    # Build some little endian type encodings
    double_type = np.dtype(np.double)
    double_type = double_type.newbyteorder('<')
    int_type = np.dtype(np.int32)
    int_type = int_type.newbyteorder('<')

    # Decode the simple buffers
    result['energies'] = np.frombuffer(base64.b64decode(result['energies']),
                                       dtype=double_type)

    # adjust energies by offset (in future this might be handled by SAPI)
    offset = result.setdefault('offset', 0)
    if offset:
        # we need to make a copy because frombuffer returns read-only array
        result['energies'] = result['energies'] + offset

    if 'num_occurrences' in result:
        result['num_occurrences'] = \
            np.frombuffer(base64.b64decode(result['num_occurrences']),
                        dtype=int_type)

    result['active_variables'] = \
        np.frombuffer(base64.b64decode(result['active_variables']),
                      dtype=int_type)

    # Measure out the binary data size
    num_solutions = len(result['energies'])
    active_variables = result['active_variables']
    num_variables = len(active_variables)
    total_variables = result['num_variables']

    # Decode the solutions, which will be a continuous run of bits
    byte_type = np.dtype(np.uint8)
    byte_type = byte_type.newbyteorder('<')
    bits = np.unpackbits(np.frombuffer(base64.b64decode(result['solutions']),
                         dtype=byte_type))

    # Clip off the extra bits from encoding
    if num_solutions:
        bits = np.reshape(bits, (num_solutions, bits.size // num_solutions))
        bits = np.delete(bits, range(num_variables, bits.shape[1]), 1)

    # Switch from bits to spins
    default = 3
    if msg['type'] == 'ising':
        bits = bits.astype(np.int8)
        bits *= 2
        bits -= 1
        default = 0

    # Fill in the missing variables
    solutions = np.full((num_solutions, total_variables), default, dtype=np.int8)
    solutions[:, active_variables] = bits
    result['solutions'] = solutions

    # If the final result shouldn't be numpy formats switch back to python objects
    if not return_matrix:
        result['energies'] = result['energies'].tolist()
        if 'num_occurrences' in result:
            result['num_occurrences'] = result['num_occurrences'].tolist()
        result['active_variables'] = result['active_variables'].tolist()
        result['solutions'] = result['solutions'].tolist()

    # include problem type
    if 'type' in msg:
        result['problem_type'] = msg['type']

    return result


def encode_problem_as_ref(problem):
    """Encode the problem given via reference for submission in the `ref` data
    format.

    Args:
        problem (str):
            A reference to an uploaded problem (problem ID).

    Returns:
        Encoded submission dictionary.
    """

    if not isinstance(problem, str):
        raise TypeError("unsupported problem reference type")

    return {
        'format': 'ref',
        'data': problem
    }


def encode_problem_as_bq(problem):
    """Encode the binary quadratic problem for submission in the `bq` data
    format.

    Args:
        problem (:class:`~dimod.BinaryQuadraticModel`):
            A binary quadratic model.

    Returns:
        Encoded submission dictionary.

    Note:
        The `bq` format assumes the complete BQM is sent embedded in the sample
        job submit data, something none of the production solvers currently
        support.
    """
    # NOTE: semi-deprecated format; see `bqm_as_file`.

    if not hasattr(problem, 'to_serializable'):
        raise TypeError("unsupported problem type")

    return {
        'format': 'bq',
        'data': problem.to_serializable(use_bytes=False)
    }


def decode_bq(msg):
    """Decode answer for problem submitted in the `bq` data format."""
    try:
        import dimod
    except ImportError:     # pragma: no cover
        raise RuntimeError("Can't decode BQMs without dimod. "
                           "Re-install the library with 'bqm' support.")

    answer = msg['answer']
    assert answer['format'] == 'bq'

    result = {}

    # sampleset is encoded in data field
    result['sampleset'] = dimod.SampleSet.from_serializable(answer['data'])

    # include problem type
    result['problem_type'] = msg['type']

    return result


def bqm_as_file(bqm, **options):
    """Encode in-memory BQM as DIMODBQM binary file format.

    Args:
        bqm (:class:`~dimod.BQM`):
            Binary quadratic model.

        **options (dict):
            :class:`~dimod.serialization.fileview.FileView` options.

    Returns:
        file-like:
            Binary stream with BQM encoded in DIMODBQM format.
    """
    # TODO: replace with `bqm.to_file` when we drop support for dimod 0.9.x

    try:
        import dimod
    except ImportError: # pragma: no cover
        raise RuntimeError("Can't encode BQM without 'dimod'. "
                           "Re-install the library with 'bqm' support.")

    try:
        # Using `bqm.to_file()` for serialization is preferred since
        # dwavesystems/dimod#599, but we need a fallback for older dimods.
        # Specifically:
        # - `Adj{Vector,Array,Map}BQM.to_file` was added in dimod 0.9.6.
        # - `BQM.to_file` method was added in dimod 0.10.0.
        bqm.to_file
    except AttributeError: # pragma: no cover
        pass
    else:
        return bqm.to_file(**options)

    try:
        # we need FileView in dimod < 0.9.6
        from dimod.serialization.fileview import FileView as BQMFileView
    except ImportError: # pragma: no cover
        # this should never happen
        raise RuntimeError(
            f"Can't import FileView serializer from dimod=={dimod.__version__}")

    if isinstance(bqm, BQMFileView):
        return bqm

    # test explicitly to avoid copy on cast if possible
    # note: it's safe to use Adj*BQMs (removed in dimod 0.10), as we never get
    # here if dimod 0.10 is used
    fileviewable = (dimod.AdjArrayBQM, dimod.AdjVectorBQM, dimod.AdjMapBQM)
    if not isinstance(bqm, fileviewable):
        bqm = dimod.AdjVectorBQM(bqm)

    return BQMFileView(bqm, **options)



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import time
import json
import unittest
import tempfile
import collections
from unittest import mock
from concurrent.futures import ThreadPoolExecutor, wait

from requests.exceptions import HTTPError
from parameterized import parameterized

from dwave.cloud.utils import tictoc
from dwave.cloud.client import Client
from dwave.cloud.exceptions import SAPIError, ProblemUploadError
from dwave.cloud.upload import (
    Gettable, GettableFile, GettableMemory, FileView, ChunkedData)

from tests import config


class TestGettableABC(unittest.TestCase):

    def test_invalid(self):
        class InvalidGettable(Gettable):
            pass

        with self.assertRaises(TypeError):
            InvalidGettable()

    def test_valid(self):
        class ValidGettable(Gettable):
            def __len__(self):
                return NotImplementedError
            def __getitem__(self, key):
                return NotImplementedError
            def getinto(self, key):
                return NotImplementedError

        try:
            ValidGettable()
        except:
            self.fail("unexpected interface of Gettable")


class TestGettables(unittest.TestCase):
    data = b'0123456789'

    def verify_getitem(self, gettable, data):
        n = len(data)
        # python 2 fix: indexing of bytes returns a slice (not int)
        data = bytearray(data)

        # integer indexing
        self.assertEqual(gettable[0], data[0])
        self.assertEqual(gettable[n-1], data[n-1])

        # negative integer indexing
        self.assertEqual(gettable[-1], data[-1])
        self.assertEqual(gettable[-n], data[-n])

        # out of bounds integer indexing
        with self.assertRaises(IndexError):
            gettable[n]

        # non-integer key
        with self.assertRaises(TypeError):
            gettable['a']

        # empty slices
        self.assertEqual(gettable[1:0], b'')

        # slicing
        self.assertEqual(gettable[:], data[:])
        self.assertEqual(gettable[0:n//2], data[0:n//2])
        self.assertEqual(gettable[-n//2:], data[-n//2:])

    def verify_getinto(self, gettable, data):
        n = len(data)
        # python 2 fix: indexing of bytes returns a slice (not int)
        data = bytearray(data)

        # integer indexing
        b = bytearray(n)
        self.assertEqual(gettable.getinto(0, b), 1)
        self.assertEqual(b[0], data[0])

        self.assertEqual(gettable.getinto(n-1, b), 1)
        self.assertEqual(b[0], data[n-1])

        # negative integer indexing
        self.assertEqual(gettable.getinto(-1, b), 1)
        self.assertEqual(b[0], data[-1])

        self.assertEqual(gettable.getinto(-n, b), 1)
        self.assertEqual(b[0], data[-n])

        # out of bounds integer indexing => nop
        self.assertEqual(gettable.getinto(n, b), 0)

        # non-integer key
        with self.assertRaises(TypeError):
            gettable.getinto('a', b)

        # empty slices
        self.assertEqual(gettable.getinto(slice(1, 0), b), 0)

        # slicing
        b = bytearray(n)
        self.assertEqual(gettable.getinto(slice(None), b), n)
        self.assertEqual(b, data)

        b = bytearray(n)
        self.assertEqual(gettable.getinto(slice(0, n//2), b), n//2)
        self.assertEqual(b[0:n//2], data[0:n//2])
        self.assertEqual(b[n//2:], bytearray(n//2))

        b = bytearray(n)
        self.assertEqual(gettable.getinto(slice(-n//2, None), b), n//2)
        self.assertEqual(b[:n//2], data[-n//2:])
        self.assertEqual(b[n//2:], bytearray(n//2))

        # slicing into a buffer too small
        m = 3
        b = bytearray(m)
        self.assertEqual(gettable.getinto(slice(None), b), m)
        self.assertEqual(b, data[:m])

    def test_gettable_file_from_memory_bytes(self):
        data = self.data
        fp = io.BytesIO(data)
        gf = GettableFile(fp)

        self.assertEqual(len(gf), len(data))
        self.verify_getitem(gf, data)
        self.verify_getinto(gf, data)

    def test_gettable_file_from_memory_string(self):
        data = self.data.decode()
        fp = io.StringIO(data)

        with self.assertRaises(TypeError):
            GettableFile(fp)

    def test_gettable_file_from_file_like(self):
        data = self.data

        # create file-like temporary object (on POSIX this is a tmp file)
        with tempfile.TemporaryFile() as fp:
            fp.write(data)
            fp.seek(0)
            gf = GettableFile(fp, strict=False)

            self.assertEqual(len(gf), len(data))
            self.verify_getitem(gf, data)
            self.verify_getinto(gf, data)

    def test_gettable_file_from_disk_file(self):
        data = self.data

        # create temporary file
        fd, path = tempfile.mkstemp()
        os.write(fd, data)
        os.close(fd)

        # test GettableFile from file on disk (read access)
        with io.open(path, 'rb') as fp:
            gf = GettableFile(fp)

            self.assertEqual(len(gf), len(data))
            self.verify_getitem(gf, data)
            self.verify_getinto(gf, data)

        # works also for read+write access
        with io.open(path, 'r+b') as fp:
            gf = GettableFile(fp)

            self.assertEqual(len(gf), len(data))
            self.verify_getitem(gf, data)
            self.verify_getinto(gf, data)

        # fail without read access
        with io.open(path, 'wb') as fp:
            with self.assertRaises(TypeError):
                GettableFile(fp)

        # remove temp file
        os.unlink(path)

    def test_gettable_file_critical_section_respected(self):
        # setup a shared file view
        data = self.data
        fp = io.BytesIO(data)
        gf = GettableFile(fp)

        # file slices
        slice_a = slice(0, 7)
        slice_b = slice(3, 5)

        # add a noticeable sleep inside the critical section (on `file.seek`),
        # resulting in minimal runtime equal to (N runs * sleep in crit sect)
        sleep = 0.25
        def blocking_seek(start):
            time.sleep(sleep)
            return io.BytesIO.seek(gf._fp, start)
        gf._fp.seek = blocking_seek

        # define the worker
        def worker(slice_):
            return gf[slice_]

        # run the worker a few times in parallel
        executor = ThreadPoolExecutor(max_workers=3)
        slices = [slice_a, slice_b, slice_a]
        futures = [executor.submit(worker, s) for s in slices]
        with tictoc() as timer:
            wait(futures)

        # verify results
        results = [f.result() for f in futures]
        expected = [data[s] for s in slices]
        self.assertEqual(results, expected)

        # verify runtime is consistent with a blocking critical section
        self.assertGreaterEqual(timer.dt, 0.9 * len(results) * sleep)

    def test_gettable_memory_from_bytes_like(self):
        data_objects = [
            bytes(self.data),
            bytearray(self.data),
            memoryview(self.data)
        ]

        for data in data_objects:
            gm = GettableMemory(data)

            self.assertEqual(len(gm), len(data))
            self.verify_getitem(gm, data)
            self.verify_getinto(gm, data)


class TestFileView(unittest.TestCase):
    # python 2 fix: indexing of bytes returns a slice (not int)
    data = bytearray(b'0123456789')

    def test_file_interface(self):
        data = self.data
        size = len(data)
        fp = io.BytesIO(data)
        gf = GettableFile(fp)
        fv = FileView(gf)

        # partial read
        self.assertEqual(fv.read(1), data[0:1])

        # read all, also check continuity
        self.assertEqual(fv.read(), data[1:])

        # seek and tell
        self.assertEqual(len(fv), size)

        self.assertEqual(fv.seek(2), 2)
        self.assertEqual(fv.tell(), 2)

        self.assertEqual(fv.seek(2, io.SEEK_CUR), 4)
        self.assertEqual(fv.tell(), 4)

        self.assertEqual(fv.seek(0, io.SEEK_END), size)
        self.assertEqual(fv.tell(), size)

        # IOBase derived methods
        fv.seek(0)
        self.assertEqual(fv.readlines(), [data])

    def test_view_interface(self):
        data = self.data
        size = len(data)
        fp = io.BytesIO(data)
        gf = GettableFile(fp)
        fv = FileView(gf)

        # view, slice index
        subfv = fv[1:-1]
        self.assertEqual(subfv.read(), data[1:-1])
        self.assertEqual(len(subfv), size - 2)

        # view, integer index
        self.assertEqual(fv[2], data[2])

        # view, out of bounds index
        with self.assertRaises(IndexError):
            fv[size]

        # view are independent
        self.assertEqual(fv[:2].read(), data[:2])
        self.assertEqual(fv[-2:].read(), data[-2:])


class TestChunkedData(unittest.TestCase):
    data = b'0123456789'

    def verify_chunking(self, cd, chunks_expected):
        self.assertEqual(len(cd), len(chunks_expected))
        self.assertEqual(cd.num_chunks, len(chunks_expected))
        self.assertEqual(cd.total_size, sum(map(len, chunks_expected)))

        chunks_iter = [c.read() for c in cd]
        chunks_explicit = []
        for idx in range(len(cd)):
            chunks_explicit.append(cd.chunk(idx).read())

        self.assertListEqual(chunks_iter, chunks_expected)
        self.assertListEqual(chunks_explicit, chunks_iter)

    def test_chunks_from_bytes(self):
        cd = ChunkedData(self.data, chunk_size=3)
        chunks_expected = [b'012', b'345', b'678', b'9']
        self.verify_chunking(cd, chunks_expected)

    def test_chunks_from_bytearray(self):
        cd = ChunkedData(bytearray(self.data), chunk_size=3)
        chunks_expected = [b'012', b'345', b'678', b'9']
        self.verify_chunking(cd, chunks_expected)

    def test_chunks_from_str(self):
        cd = ChunkedData(self.data.decode('ascii'), chunk_size=3)
        chunks_expected = [b'012', b'345', b'678', b'9']
        self.verify_chunking(cd, chunks_expected)

    def test_chunks_from_memory_file(self):
        data = io.BytesIO(self.data)
        cd = ChunkedData(data, chunk_size=3)
        chunks_expected = [b'012', b'345', b'678', b'9']
        self.verify_chunking(cd, chunks_expected)

    @parameterized.expand([
        ("BQM",),
        ("BinaryQuadraticModel",),
        ("AdjVectorBQM",),
    ])
    def test_chunks_from_bqm(self, bqm_cls_name):
        try:
            import dimod
            bqm_cls = getattr(dimod, bqm_cls_name)
            bqm_cls.to_file
        except (ImportError, AttributeError):
            self.skipTest(f"dimod.{bqm_cls_name}.to_file() unavailable")

        # serialize a BQM via .to_file
        bqm = bqm_cls.from_ising({'a': 1}, {})
        bqm_file = bqm.to_file()     # returns dimod's FileView

        chunk_size = 10
        cd = ChunkedData(bqm_file, chunk_size=chunk_size)

        # verify chunks
        bqm_file.seek(0)
        raw = bqm_file.read()
        chunks_expected = [raw[i:i+chunk_size] for i in range(0, len(raw), chunk_size)]
        self.verify_chunking(cd, chunks_expected)

    def test_chunks_from_dqm(self):
        try:
            from dimod import DQM
        except ImportError:
            self.skipTest("dimod.DQM unavailable")

        # serialize a DQM via .to_file
        dqm = DQM()
        dqm.add_variable(1)
        dqm_file = dqm.to_file()    # returns SpooledTemporaryFile subclass

        chunk_size = 100
        cd = ChunkedData(dqm_file, chunk_size=chunk_size)

        # verify chunks
        dqm_file.seek(0)
        raw = dqm_file.read()
        chunks_expected = [raw[i:i+chunk_size] for i in range(0, len(raw), chunk_size)]
        self.verify_chunking(cd, chunks_expected)

    def test_chunk_size_edges(self):
        with self.assertRaises(ValueError):
            cd = ChunkedData(self.data, chunk_size=0)

        cd = ChunkedData(self.data, chunk_size=1)
        chunks_expected = [self.data[i:i+1] for i in range(len(self.data))]
        self.verify_chunking(cd, chunks_expected)

        cd = ChunkedData(self.data, chunk_size=len(self.data))
        chunks_expected = [self.data]
        self.verify_chunking(cd, chunks_expected)

    def test_chunk_generators(self):
        cd = ChunkedData(self.data, chunk_size=3)
        chunks_expected = [b'012', b'345', b'678', b'9']
        chunks_generated = [g().read() for g in cd.generators()]
        self.assertListEqual(chunks_expected, chunks_generated)


def choose_reply(key, replies, statuses=None):
    """Choose the right response based on a hashable `key` and make a mock
    response.
    """

    if statuses is None:
        statuses = collections.defaultdict(lambda: iter([200]))

    if key in replies:
        response = mock.Mock(['text', 'json', 'raise_for_status'])
        response.status_code = next(statuses[key])
        response.text = replies[key]
        response.json.side_effect = lambda: json.loads(replies[key])

        def raise_for_status():
            if not 200 <= response.status_code < 400:
                raise HTTPError(response.status_code)
        response.raise_for_status = raise_for_status

        def ok():
            try:
                response.raise_for_status()
            except HTTPError:
                return False
            return True
        ok_property = mock.PropertyMock(side_effect=ok)
        type(response).ok = ok_property

        return response
    else:
        raise NotImplementedError(key)


@mock.patch('time.sleep', lambda *args: None)
class TestMockedMultipartUpload(unittest.TestCase):

    @mock.patch.multiple(Client, _UPLOAD_PART_SIZE_BYTES=1)
    def test_single_problem_end_to_end(self):
        """Verify a fresh problem multipart upload works end to end."""

        upload_data = b'123'
        upload_problem_id = '84ef154c-28f9-46ed-9f22-aec0583499f2'

        parts = list(range(len(upload_data)))
        part_data = [upload_data[i:i+1] for i in parts]

        _md5 = Client._digest
        _hex = Client._checksum_hex
        _b64 = Client._checksum_b64
        part_digest = [_md5(part_data[i]) for i in parts]
        combine_checksum = _hex(_md5(b''.join(part_digest)))

        # we need a "global session", because mocked responses are stateful
        def global_mock_session():
            session = mock.MagicMock()
            session.__enter__ = lambda *args: session

            def get(path, seq=iter(range(2))):
                all_parts = [{"part_number": i+1,
                              "checksum": _hex(part_digest[i])} for i in parts]

                return choose_reply((path, next(seq)), {
                    # initial upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 0):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": []}),

                    # final upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 1):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": all_parts}),
                })

            def post(path, **kwargs):
                json_ = kwargs.pop('json')
                body = json.dumps(sorted(json_.items()))
                return choose_reply((path, body), {
                    # initiate upload
                    ('bqm/multipart',
                     json.dumps([('size', len(upload_data))])):
                        json.dumps({'id': upload_problem_id}),

                    # combine parts
                    ('bqm/multipart/{}/combine'.format(upload_problem_id),
                     json.dumps([('checksum', combine_checksum)])):
                        json.dumps({}),
                })

            def put(path, data, headers):
                body = data.read()
                headers = json.dumps(sorted(headers.items()))
                replies = {
                    (
                        'bqm/multipart/{}/part/{}'.format(upload_problem_id, i+1),
                        part_data[i],
                        json.dumps(sorted([
                            ('Content-MD5', _b64(part_digest[i])),
                            ('Content-Type', 'application/octet-stream')
                        ]))
                    ): json.dumps({})
                    for i in parts
                }
                return choose_reply((path, body, headers), replies)

            session.get = get
            session.put = put
            session.post = post

            return session

        session = global_mock_session()

        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:

                future = client.upload_problem_encoded(upload_data)
                try:
                    returned_problem_id = future.result()
                except Exception as e:
                    self.fail(e)

                self.assertEqual(returned_problem_id, upload_problem_id)

    @mock.patch.multiple(Client, _UPLOAD_PART_SIZE_BYTES=1)
    def test_partial_upload(self):
        """Verify only missing parts are uploaded."""

        upload_data = b'123'
        upload_problem_id = '84ef154c-28f9-46ed-9f22-aec0583499f2'

        parts = list(range(len(upload_data)))
        part_data = [upload_data[i:i+1] for i in parts]

        _md5 = Client._digest
        _hex = Client._checksum_hex
        _b64 = Client._checksum_b64
        part_digest = [_md5(part_data[i]) for i in parts]
        combine_checksum = _hex(_md5(b''.join(part_digest)))

        # we need a "global session", because mocked responses are stateful
        def global_mock_session():
            session = mock.MagicMock()
            session.__enter__ = lambda *args: session

            def get(path, seq=iter(range(2))):
                all_parts = [{"part_number": i+1,
                              "checksum": _hex(part_digest[i])} for i in parts]

                return choose_reply((path, next(seq)), {
                    # initial upload status: all parts uploaded except the first one
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 0):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": all_parts[1:]}),

                    # final upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 1):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": all_parts}),
                })

            def post(path, **kwargs):
                json_ = kwargs.pop('json')
                body = json.dumps(sorted(json_.items()))
                return choose_reply((path, body), {
                    # initiate upload
                    ('bqm/multipart',
                     json.dumps([('size', len(upload_data))])):
                        json.dumps({'id': upload_problem_id}),

                    # combine parts
                    ('bqm/multipart/{}/combine'.format(upload_problem_id),
                     json.dumps([('checksum', combine_checksum)])):
                        json.dumps({}),
                })

            def put(path, data, headers):
                body = data.read()
                headers = json.dumps(sorted(headers.items()))
                replies = {
                    # only the first part!
                    (
                        'bqm/multipart/{}/part/{}'.format(upload_problem_id, i+1),
                        part_data[i],
                        json.dumps(sorted([
                            ('Content-MD5', _b64(part_digest[i])),
                            ('Content-Type', 'application/octet-stream')
                        ]))
                    ): json.dumps({})
                    for i in parts[:1]
                }
                return choose_reply((path, body, headers), replies)

            session.get = get
            session.put = put
            session.post = post

            return session

        session = global_mock_session()

        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:

                future = client.upload_problem_encoded(upload_data)
                try:
                    returned_problem_id = future.result()
                except Exception as e:
                    self.fail(e)

                self.assertEqual(returned_problem_id, upload_problem_id)

    def test_part_upload_retried(self):
        """Verify upload successful even if part upload fails a few times."""

        # using the default part size here (5MB), so we have only one part
        upload_data = b'123'
        upload_problem_id = '84ef154c-28f9-46ed-9f22-aec0583499f2'

        parts = [0]
        part_data = [upload_data]

        _md5 = Client._digest
        _hex = Client._checksum_hex
        _b64 = Client._checksum_b64
        part_digest = [_md5(part_data[i]) for i in parts]
        combine_checksum = _hex(_md5(b''.join(part_digest)))

        # we need a "global session", because mocked responses are stateful
        def global_mock_session(n_failures):
            session = mock.MagicMock()
            session.__enter__ = lambda *args: session

            def get(path, seq=iter(range(2))):
                all_parts = [{"part_number": i+1,
                              "checksum": _hex(part_digest[i])} for i in parts]

                return choose_reply((path, next(seq)), {
                    # initial upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 0):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": []}),

                    # final upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 1):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": all_parts}),
                })

            def post(path, **kwargs):
                json_ = kwargs.pop('json')
                body = json.dumps(sorted(json_.items()))
                return choose_reply((path, body), {
                    # initiate upload
                    ('bqm/multipart',
                     json.dumps([('size', len(upload_data))])):
                        json.dumps({'id': upload_problem_id}),

                    # combine parts
                    ('bqm/multipart/{}/combine'.format(upload_problem_id),
                     json.dumps([('checksum', combine_checksum)])):
                        json.dumps({}),
                })

            def put(path, data, headers, seq=iter(range(Client._UPLOAD_PART_RETRIES+1))):
                body = data.read()
                headers = json.dumps(sorted(headers.items()))
                keys = [
                    (
                        'bqm/multipart/{}/part/{}'.format(upload_problem_id, i+1),
                        part_data[i],
                        json.dumps(sorted([
                            ('Content-MD5', _b64(part_digest[i])),
                            ('Content-Type', 'application/octet-stream')
                        ]))
                    ) for i in parts
                ]
                attempt = next(seq)
                if attempt < n_failures:
                    return choose_reply((path, body, headers),
                                        replies={key: '{}' for key in keys},
                                        statuses={key: iter([500]) for key in keys})
                else:
                    return choose_reply((path, body, headers),
                                        replies={key: '{}' for key in keys})

            session.get = get
            session.put = put
            session.post = post

            return session

        # part upload fails exactly _UPLOAD_PART_RETRIES times;
        # problem upload must recover
        session = global_mock_session(n_failures=Client._UPLOAD_PART_RETRIES)
        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:

                future = client.upload_problem_encoded(upload_data)
                try:
                    returned_problem_id = future.result()
                except Exception as e:
                    self.fail(e)

                self.assertEqual(returned_problem_id, upload_problem_id)

        # part upload fails exactly _UPLOAD_PART_RETRIES + 1 times;
        # problem upload will also fail
        session = global_mock_session(n_failures=Client._UPLOAD_PART_RETRIES + 1)
        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:

                with self.assertRaises(ProblemUploadError):
                    client.upload_problem_encoded(upload_data).result()

    @mock.patch.multiple(Client, _UPLOAD_PART_SIZE_BYTES=1)
    def test_problem_reupload_end_to_end(self):
        """Verify problem multipart upload continued."""

        upload_data = b'123'
        upload_problem_id = '84ef154c-28f9-46ed-9f22-aec0583499f2'

        parts = list(range(len(upload_data)))
        part_data = [upload_data[i:i+1] for i in parts]

        _md5 = Client._digest
        _hex = Client._checksum_hex
        _b64 = Client._checksum_b64
        part_digest = [_md5(part_data[i]) for i in parts]
        combine_checksum = _hex(_md5(b''.join(part_digest)))

        # we need a "global session", because mocked responses are stateful
        def global_mock_session():
            session = mock.MagicMock()
            session.__enter__ = lambda *args: session

            def get(path, seq=iter(range(2))):
                all_parts = [{"part_number": i+1,
                              "checksum": _hex(part_digest[i])} for i in parts]

                return choose_reply((path, next(seq)), {
                    # initial upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 0):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": all_parts[:2]}),

                    # final upload status
                    ('bqm/multipart/{}/status'.format(upload_problem_id), 1):
                        json.dumps({"status": "UPLOAD_IN_PROGRESS", "parts": all_parts}),
                })

            def post(path, **kwargs):
                json_ = kwargs.pop('json')
                body = json.dumps(sorted(json_.items()))
                return choose_reply((path, body), {
                    # combine parts
                    ('bqm/multipart/{}/combine'.format(upload_problem_id),
                     json.dumps([('checksum', combine_checksum)])):
                        json.dumps({}),
                })

            def put(path, data, headers):
                body = data.read()
                headers = json.dumps(sorted(headers.items()))
                replies = {
                    (
                        'bqm/multipart/{}/part/{}'.format(upload_problem_id, i+1),
                        part_data[i],
                        json.dumps(sorted([
                            ('Content-MD5', _b64(part_digest[i])),
                            ('Content-Type', 'application/octet-stream')
                        ]))
                    ): json.dumps({})
                    for i in parts[2:]
                }
                return choose_reply((path, body, headers), replies)

            session.get = get
            session.put = put
            session.post = post

            return session

        session = global_mock_session()

        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:

                future = client.upload_problem_encoded(
                    upload_data, problem_id=upload_problem_id)

                try:
                    returned_problem_id = future.result()
                except Exception as e:
                    self.fail(e)

                self.assertEqual(returned_problem_id, upload_problem_id)


@unittest.skipUnless(config, "No live server configuration available.")
class TestMultipartUpload(unittest.TestCase):
    _100gb = 100 * 2**30

    def test_smoke_test(self):
        data = b'123'
        with Client(**config) as client:
            future = client.upload_problem_encoded(data)
            try:
                problem_id = future.result()
            except Exception as e:
                self.fail(e)

    def test_initiate_size_limit(self):
        size = self._100gb

        with Client(**config) as client:
            with client.create_session() as session:
                with self.assertRaisesRegex(SAPIError,
                                            'bigger than the maximum'):
                    client._initiate_multipart_upload(session, size)

    @mock.patch.object(ChunkedData, 'total_size', _100gb)
    def test_initiate_size_limit_end_to_end(self):

        with Client(**config) as client:
            future = client.upload_problem_encoded(b'')
            with self.assertRaisesRegex(ProblemUploadError,
                                        'bigger than the maximum'):
                problem_id = future.result()



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest
import configparser
from unittest import mock
from functools import partial

from dwave.cloud.package_info import __packagename__, __version__
from dwave.cloud.exceptions import ConfigFileParseError, ConfigFileReadError
from dwave.cloud.testing import iterable_mock_open
from dwave.cloud.config import (
    get_configfile_paths, load_config_from_files, load_config,
    parse_float, parse_int, parse_boolean, get_cache_dir, update_config)


class TestConfigParsing(unittest.TestCase):

    config_body = """
        [defaults]
        endpoint = https://cloud.dwavesys.com/sapi
        client = qpu
        profile = software

        [dw2000]
        solver = DW_2000Q_1
        token = ...

        [software]
        client = sw
        solver = c4-sw_sample
        token = ...

        [alpha]
        endpoint = https://url.to.alpha/api
        proxy = http://user:pass@myproxy.com:8080/
        token = alpha-token
        headers =  key-1:value-1
          key-2: value-2
    """

    def parse_config_string(self, text):
        config = configparser.ConfigParser(default_section="defaults")
        config.read_string(text)
        return config

    def test_config_load_from_file(self):
        with mock.patch('dwave.cloud.config.open', iterable_mock_open(self.config_body), create=True):
            config = load_config_from_files(filenames=["filename"])
            self.assertEqual(config.sections(), ['dw2000', 'software', 'alpha'])
            self.assertEqual(config['dw2000']['client'], 'qpu')
            self.assertEqual(config['software']['client'], 'sw')

    def setUp(self):
        # clear `config_load`-relevant environment variables before testing, so
        # we only need to patch the ones that we are currently testing
        for key in frozenset(os.environ.keys()):
            if key.startswith("DWAVE_") or key.startswith("DW_INTERNAL__"):
                os.environ.pop(key, None)

    def test_config_load_from_file__invalid_format__duplicate_sections(self):
        """Config loading should fail with ``ConfigFileParseError`` for invalid
        config files."""
        myconfig = """
            [section]
            key = val
            [section]
            key = val
        """
        with mock.patch('dwave.cloud.config.open', iterable_mock_open(myconfig), create=True):
            self.assertRaises(ConfigFileParseError, load_config_from_files, filenames=["filename"])
            self.assertRaises(ConfigFileParseError, load_config, config_file="filename", profile="section")

    def test_no_config_detected(self):
        """When no config file detected, `load_config_from_files` should return
        empty config."""
        with mock.patch("dwave.cloud.config.get_configfile_paths", lambda: []):
            self.assertFalse(load_config_from_files().sections())

    def test_invalid_filename_given(self):
        self.assertRaises(ConfigFileReadError, load_config_from_files, filenames=['/path/to/non/existing/config'])

    def test_config_file_detection_cwd(self):
        configpath = os.path.join('.', 'dwave.conf')
        with mock.patch("os.path.exists", lambda path: path == configpath):
            self.assertEqual(get_configfile_paths(), [configpath])

    def test_config_file_detection_user(self):
        if sys.platform == 'win32':
            configpath = os.path.expanduser("~\\AppData\\Local\\dwavesystem\\dwave\\dwave.conf")
        elif sys.platform == 'darwin':
            configpath = os.path.expanduser("~/Library/Application Support/dwave/dwave.conf")
        else:
            configpath = os.path.expanduser("~/.config/dwave/dwave.conf")

        with mock.patch("os.path.exists", lambda path: path == configpath):
            self.assertEqual(get_configfile_paths(), [configpath])

    def test_config_file_detection_system(self):
        if sys.platform == 'win32':
            configpath = os.path.expandvars("%SYSTEMDRIVE%\\ProgramData\\dwavesystem\\dwave\\dwave.conf")
        elif sys.platform == 'darwin':
            configpath = "/Library/Application Support/dwave/dwave.conf"
        else:
            configpath = "/etc/xdg/dwave/dwave.conf"

        with mock.patch("os.path.exists", lambda path: path == configpath):
            self.assertEqual(get_configfile_paths(), [configpath])

    def test_config_file_detection_nonexisting(self):
        with mock.patch("os.path.exists", lambda path: False):
            self.assertEqual(get_configfile_paths(), [])

    def test_config_file_path_expansion(self):
        """Home dir and env vars are expanded when resolving config path."""

        env = {"var": "val"}
        config_file = "~/path/${var}/to/$var/my.conf"
        expected_path = os.path.expanduser("~/path/val/to/val/my.conf")
        profile = "profile"

        conf_content = """
            [{}]
            valid = yes
        """.format(profile)

        def mock_open(filename, *pa, **kw):
            self.assertEqual(filename, expected_path)
            return iterable_mock_open(conf_content)()

        # config file via kwarg
        with mock.patch.dict(os.environ, env):
            with mock.patch('dwave.cloud.config.open', mock_open) as m:
                conf = load_config(config_file=config_file, profile=profile)
                self.assertEqual(conf['valid'], 'yes')

        # config file via env var
        with mock.patch.dict(os.environ, env, DWAVE_CONFIG_FILE=config_file):
            with mock.patch('dwave.cloud.config.open', mock_open) as m:
                conf = load_config(profile=profile)
                self.assertEqual(conf['valid'], 'yes')

    def _assert_config_valid(self, config):
        # profile 'alpha' is loaded
        self.assertEqual(config['endpoint'], "https://url.to.alpha/api")
        # default values are inherited
        self.assertEqual(config['client'], "qpu")
        # multi-line values are read
        self.assertEqual(config['headers'], "key-1:value-1\nkey-2: value-2")

    def _load_config_from_files(self, asked, provided=None, data=None):
        self.assertEqual(asked, provided)
        if data is None:
            data = self.config_body
        return self.parse_config_string(data)


    def test_config_load_configfile_arg(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=['myfile'])):
            self._assert_config_valid(load_config(config_file='myfile', profile='alpha'))

    def test_config_load_configfile_env(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=['myfile'])):
            with mock.patch.dict(os.environ, {'DWAVE_CONFIG_FILE': 'myfile'}):
                self._assert_config_valid(load_config(config_file=None, profile='alpha'))

    def test_config_load_configfile_detect(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=None)):
            self._assert_config_valid(load_config(config_file=None, profile='alpha'))

    def test_config_load_skip_configfiles(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        self._load_config_from_files):

            # don't load from file, use arg or env
            self.assertNotIn('endpoint', load_config(config_file=False))
            with mock.patch.dict(os.environ, {'DWAVE_API_ENDPOINT': 'test'}):
                self.assertEqual(load_config(config_file=False)['endpoint'], 'test')

            # specifying a profile doesn't affect outcome
            self.assertNotIn('endpoint', load_config(config_file=False, profile='alpha'))
            with mock.patch.dict(os.environ, {'DWAVE_API_ENDPOINT': 'test'}):
                self.assertEqual(load_config(config_file=False, profile='alpha')['endpoint'], 'test')
            with mock.patch.dict(os.environ, {'DWAVE_PROFILE': 'profile'}):
                self.assertEqual(load_config(config_file=False, endpoint='test')['endpoint'], 'test')

    def test_config_load_force_autodetection(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=None)):

            # load from file
            self._assert_config_valid(load_config(config_file=True, profile='alpha'))

            # load from file, even when config_file overridden in env (to path)
            with mock.patch.dict(os.environ, {'DWAVE_CONFIG_FILE': 'nonexisting'}):
                self._assert_config_valid(load_config(config_file=True, profile='alpha'))
                with mock.patch.dict(os.environ, {'DWAVE_PROFILE': 'alpha'}):
                    self._assert_config_valid(load_config(config_file=True))

            # load from file, even when config_file overridden in env (to None)
            with mock.patch.dict(os.environ, {'DWAVE_CONFIG_FILE': ''}):
                self._assert_config_valid(load_config(config_file=True, profile='alpha'))
                with mock.patch.dict(os.environ, {'DWAVE_PROFILE': 'alpha'}):
                    self._assert_config_valid(load_config(config_file=True))

    def test_config_load_configfile_detect_profile_env(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=None)):
            with mock.patch.dict(os.environ, {'DWAVE_PROFILE': 'alpha'}):
                self._assert_config_valid(load_config())

    def test_config_load_configfile_env_profile_env(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=['myfile'])):
            with mock.patch.dict(os.environ, {'DWAVE_CONFIG_FILE': 'myfile',
                                              'DWAVE_PROFILE': 'alpha'}):
                self._assert_config_valid(load_config())

    def test_config_load_configfile_env_profile_env_key_arg(self):
        """Explicitly provided values should override env/file."""
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=['myfile'])):
            with mock.patch.dict(os.environ, {'DWAVE_CONFIG_FILE': 'myfile',
                                              'DWAVE_PROFILE': 'alpha'}):
                self.assertEqual(load_config(endpoint='manual')['endpoint'], 'manual')
                self.assertEqual(load_config(token='manual')['token'], 'manual')
                self.assertEqual(load_config(client='manual')['client'], 'manual')
                self.assertEqual(load_config(solver='manual')['solver'], 'manual')
                self.assertEqual(load_config(proxy='manual')['proxy'], 'manual')
                self.assertEqual(load_config(headers='headers')['headers'], 'headers')

    def test_config_load__profile_arg_nonexisting(self):
        """load_config should fail if the profile specified in kwargs or env in
        non-existing.
        """
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=None)):
            self.assertRaises(ValueError, load_config, profile="nonexisting")
            with mock.patch.dict(os.environ, {'DWAVE_PROFILE': 'nonexisting'}):
                self.assertRaises(ValueError, load_config)

    def test_config_load_configfile_arg_profile_default(self):
        """Check the right profile is loaded when `profile` specified only in
        [defaults] section.
        """
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, provided=['myfile'])):
            profile = load_config(config_file='myfile')
            self.assertEqual(profile['solver'], 'c4-sw_sample')

    def test_config_load__profile_first_section(self):
        """load_config should load the first section for profile, if profile
        is nowhere else specified.
        """
        myconfig = """
            [first]
            solver = DW_2000Q_1
        """
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files,
                                provided=None, data=myconfig)):
            profile = load_config()
            self.assertIn('solver', profile)
            self.assertEqual(profile['solver'], 'DW_2000Q_1')

    def test_config_load__profile_from_defaults(self):
        """load_config should promote [defaults] section to profile, if profile
        is nowhere else specified *and* not even a single non-[defaults] section
        exists.
        """
        myconfig = """
            [defaults]
            solver = DW_2000Q_1
        """
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files,
                                provided=None, data=myconfig)):
            profile = load_config()
            self.assertIn('solver', profile)
            self.assertEqual(profile['solver'], 'DW_2000Q_1')

    def test_config_load_configfile_arg_profile_default_nonexisting(self):
        """load_config should fail if the profile specified in the defaults
        section is non-existing.
        """
        myconfig = """
            [defaults]
            profile = nonexisting

            [some]
            solver = DW_2000Q_1
        """
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files,
                                provided=['myfile'], data=myconfig)):
            self.assertRaises(ValueError, load_config, config_file='myfile')

    def test_config_load_multiple_autodetected_configfiles(self):
        """Test more specific config overrides less specific one,
        on a key by key basis, in a list of auto-detected config files."""

        config_system = """
            [alpha]
            endpoint = alpha
            solver = DW_2000Q_1
        """
        config_user = """
            [alpha]
            solver = DW_2000Q_2
            [beta]
            endpoint = beta
        """

        with mock.patch("dwave.cloud.config.get_configfile_paths",
                        lambda: ['config_system', 'config_user']):

            # test per-key override
            with mock.patch('dwave.cloud.config.open', create=True) as m:
                m.side_effect = [iterable_mock_open(config_system)(),
                                 iterable_mock_open(config_user)()]
                section = load_config(profile='alpha')
                self.assertEqual(section['endpoint'], 'alpha')
                self.assertEqual(section['solver'], 'DW_2000Q_2')

            # test per-section override (section addition)
            with mock.patch('dwave.cloud.config.open', create=True) as m:
                m.side_effect = [iterable_mock_open(config_system)(),
                                 iterable_mock_open(config_user)()]
                section = load_config(profile='beta')
                self.assertEqual(section['endpoint'], 'beta')

    def test_config_load_multiple_explicit_configfiles(self):
        """Test more specific config overrides less specific one,
        on a key by key basis, in a list of explicitly given files."""

        file1 = """
            [alpha]
            endpoint = alpha
            solver = DW_2000Q_1
        """
        file2 = """
            [alpha]
            solver = DW_2000Q_2
        """

        with mock.patch('dwave.cloud.config.open', create=True) as m:
            m.side_effect=[iterable_mock_open(file1)(),
                           iterable_mock_open(file2)()]
            section = load_config(config_file=['file1', 'file2'], profile='alpha')
            m.assert_has_calls([mock.call('file1', 'r'), mock.call('file2', 'r')])
            self.assertEqual(section['endpoint'], 'alpha')
            self.assertEqual(section['solver'], 'DW_2000Q_2')

    def test_config_load_env_override(self):
        with mock.patch("dwave.cloud.config.load_config_from_files",
                        partial(self._load_config_from_files, data="", provided=['myfile'])):

            with mock.patch.dict(os.environ, {'DWAVE_API_CLIENT': 'test'}):
                self.assertEqual(load_config(config_file='myfile')['client'], 'test')

            with mock.patch.dict(os.environ, {'DWAVE_API_ENDPOINT': 'test'}):
                self.assertEqual(load_config(config_file='myfile')['endpoint'], 'test')

            with mock.patch.dict(os.environ, {'DWAVE_API_TOKEN': 'test'}):
                self.assertEqual(load_config(config_file='myfile')['token'], 'test')

            with mock.patch.dict(os.environ, {'DWAVE_API_SOLVER': 'test'}):
                self.assertEqual(load_config(config_file='myfile')['solver'], 'test')

            with mock.patch.dict(os.environ, {'DWAVE_API_PROXY': 'test'}):
                self.assertEqual(load_config(config_file='myfile')['proxy'], 'test')

            with mock.patch.dict(os.environ, {'DWAVE_API_HEADERS': 'test'}):
                self.assertEqual(load_config(config_file='myfile')['headers'], 'test')


class TestConfigUtils(unittest.TestCase):

    def test_parse_float(self):
        self.assertEqual(parse_float(None), None)
        self.assertEqual(parse_float(''), None)
        self.assertEqual(parse_float('', default=1), 1)

        self.assertEqual(parse_float('1.5'), 1.5)
        self.assertEqual(parse_float(1.5), 1.5)
        self.assertEqual(parse_float(1), 1.0)

    def test_parse_int(self):
        self.assertEqual(parse_int(None), None)
        self.assertEqual(parse_int(''), None)
        self.assertEqual(parse_int('', default=1), 1)

        with self.assertRaises(ValueError):
            parse_int('1.5')
        with self.assertRaises(ValueError):
            parse_int(1.5)

        self.assertEqual(parse_int(123), 123)
        self.assertEqual(parse_int(0), 0)
        self.assertEqual(parse_int(-123), -123)

    def test_parse_boolean(self):
        self.assertEqual(parse_boolean(None), None)
        self.assertEqual(parse_boolean(''), None)
        self.assertEqual(parse_boolean('', default=1), 1)

        self.assertEqual(parse_boolean(True), True)
        self.assertEqual(parse_boolean('on'), True)
        self.assertEqual(parse_boolean('On'), True)
        self.assertEqual(parse_boolean('true'), True)
        self.assertEqual(parse_boolean('True'), True)
        self.assertEqual(parse_boolean('1'), True)
        self.assertEqual(parse_boolean('123'), True)
        self.assertEqual(parse_boolean('{"a": 1}'), True)

        self.assertEqual(parse_boolean(False), False)
        self.assertEqual(parse_boolean('off'), False)
        self.assertEqual(parse_boolean('Off'), False)
        self.assertEqual(parse_boolean('false'), False)
        self.assertEqual(parse_boolean('False'), False)
        self.assertEqual(parse_boolean('0'), False)
        self.assertEqual(parse_boolean('{}'), False)

        with self.assertRaises(ValueError):
            parse_boolean('x')

    def test_cache_dir(self):
        path = get_cache_dir()
        self.assertTrue(os.path.isdir(path))
        self.assertIn(__packagename__, path)
        self.assertIn(__version__, path)

    def test_config_update_simple_override(self):
        config = {'token': 1}
        update = {'token': 2, 'solver': 2}
        update_config(config, update)
        self.assertEqual(config, {'token': 2, 'solver': 2})

        config = {'endpoint': 1}
        update = {'endpoint': 2}
        update_config(config, update)
        self.assertEqual(config, {'endpoint': 2})

        config = {'region': 1}
        update = {'region': 2}
        update_config(config, update)
        self.assertEqual(config, {'region': 2})

    def test_config_update_mutually_exclusive_on_different_levels(self):
        # region/endpoint on higher level override the lower level
        config = {'endpoint': 1}
        update = {'region': 2}
        update_config(config, update)
        self.assertEqual(config, {'region': 2})

        config = {'region': 1}
        update = {'endpoint': 2}
        update_config(config, update)
        self.assertEqual(config, {'endpoint': 2})

        config = {'endpoint': 1, 'region': 1}
        update = {'region': 2}
        update_config(config, update)
        self.assertEqual(config, {'region': 2})

        config = {'endpoint': 1, 'region': 1}
        update = {'endpoint': 2}
        update_config(config, update)
        self.assertEqual(config, {'endpoint': 2})

    def test_config_update_mutually_exclusive_on_same_level(self):
        # config update should not conflate same-level mutually exclusive options
        # updates should be minimal, only when options are in direct conflict
        config = {'endpoint': 1}
        update = {'endpoint': 2, 'region': 2}
        update_config(config, update)
        self.assertEqual(config, update)

        config = {'region': 1}
        update = {'endpoint': 2, 'region': 2}
        update_config(config, update)
        self.assertEqual(config, update)

        config = {'endpoint': 1, 'region': 1}
        update = {'endpoint': 2, 'region': 2}
        update_config(config, update)
        self.assertEqual(config, update)

        config = {}
        update = {'endpoint': 2, 'region': 2}
        update_config(config, update)
        self.assertEqual(config, update)
